import numpy as np
from . import format as cc_format


def values_are_not_repeated(dictionare, name_dictionary=None):
    keys, values = cc_format.dict_to_key_array_and_value_array(dictionare)
    for v in values:
        if (v == values).sum() > 1:
            keys_equal = keys[v == values]
            if name_dictionary is None:
                name_dictionary = 'dictionare'
            template_names_keys = name_dictionary + '[\'{}\']'
            raise ValueError(
                'The following assumption is not met:\n'
                '\t' + ' \u2260 '.join([template_names_keys.format(k) for k in keys_equal]))


def keys_necessary_known_and_values_necessary_known_exist_and_other_keys_and_values_do_not_exist(
        dictionary, keys_necessary_known, values_necessary_known, name_dictionary=None):

    if not isinstance(dictionary, dict):
        if name_dictionary is None:
            raise TypeError('The type of dictionary must be dict')
        else:
            raise TypeError('The type of {} must be dict'.format(name_dictionary))

    if not isinstance(keys_necessary_known, np.ndarray):
        if isinstance(keys_necessary_known, list) or isinstance(keys_necessary_known, tuple):
            keys_necessary_known = np.asarray(keys_necessary_known, dtype=str)
        elif isinstance(keys_necessary_known, dict) or isinstance(keys_necessary_known, set):
            raise TypeError('The type of keys_necessary_known anything, but dict or set')
        else:
            keys_necessary_known = np.asarray([keys_necessary_known], dtype=str)

    if not isinstance(values_necessary_known, np.ndarray):
        if isinstance(values_necessary_known, list) or isinstance(values_necessary_known, tuple):
            values_necessary_known = np.asarray(values_necessary_known, dtype=str)
        elif isinstance(values_necessary_known, dict) or isinstance(values_necessary_known, set):
            raise TypeError('The type of values_necessary_known anything, but dict or set')
        else:
            values_necessary_known = np.asarray([values_necessary_known], dtype=str)

    if name_dictionary is None:
        message_1 = 'In the dictionary, '
    else:
        message_1 = 'In the dictionary ' + name_dictionary + ', '

    # string_keys_necessary_known = '\', \''.join(keys_necessary_known) + '.'
    message_unknown_keys_2 = 'the Key {} is unknown.\n'
    message_unknown_keys_3 = ('the following Keys are known:\n'
                              '\t{}.'.format(keys_necessary_known))
    message_unknown_keys = (message_1 + message_unknown_keys_2 +
                            message_1 + message_unknown_keys_3)

    message_missing_keys_2 = 'the following Keys are missing:\n\t{}.\n'
    message_missing_keys_3 = ('the following Keys are necessary:\n'
                              '\t{}.'.format(keys_necessary_known))
    message_missing_keys = (message_1 + message_missing_keys_2 +
                            message_1 + message_missing_keys_3)

    # string_values_necessary_known = '\', \''.join(values_necessary_known.astype(dtype=str)) + '.'
    message_unknown_values_2 = 'the Value {} is unknown.\n'
    message_unknown_values_3 = ('the following Values are known:\n'
                                '\t{}.'.format(values_necessary_known))
    message_unknown_values = (message_1 + message_unknown_values_2 +
                              message_1 + message_unknown_values_3)

    message_missing_values_2 = 'the following Values are missing:\n\t{}.\n'
    message_missing_values_3 = ('the following Values are necessary:\n'
                                '\t{}.'.format(values_necessary_known))
    message_missing_values = (message_1 + message_missing_values_2 +
                              message_1 + message_missing_values_3)
    # print('message_unknown_keys\n', message_unknown_keys)
    # print('message_missing_keys\n', message_missing_keys)
    # print('message_unknown_values\n', message_unknown_values)
    # print('message_missing_values:\n', message_missing_values)

    n_keys = len(keys_necessary_known)
    keys_missing = np.empty(n_keys, dtype=bool)
    keys_missing[:] = True

    n_values = len(values_necessary_known)
    values_missing = np.empty(n_values, dtype=bool)
    values_missing[:] = True
    for k, v in dictionary.items():

        if k in keys_necessary_known:
            keys_missing[k == keys_necessary_known] = False
        else:
            raise ValueError(message_unknown_keys.format(k))

        if v in values_necessary_known:
            values_missing[v == values_necessary_known] = False
        else:
            raise ValueError(message_unknown_values.format(v))

    if keys_missing.any():
        raise ValueError(message_missing_keys.format(keys_necessary_known[keys_missing]))
    if values_missing.any():
        raise ValueError(message_missing_values.format(values_necessary_known[values_missing]))



import numpy as np
import csv
import os
import glob
from . import array as cc_array
from . import directory as cc_directory
from . import format as cc_format
from . import combinations as cc_combinations


def lines_to_csv_file(lines, directory, headers=None):
    with open(directory, 'w', newline='') as csv_file:
        writer = csv.writer(csv_file, delimiter=',')
        if headers is not None:
            writer.writerow(headers)
        writer.writerows(lines)


def array_to_csv_files(array_in, axis_rows, axis_columns, conditions_of_directories, headers=None):

    if axis_rows > axis_columns:
        array = np.swapaxes(array_in, axis_rows, axis_columns)
        axis_rows, axis_columns = axis_columns, axis_rows
    elif axis_rows < axis_columns:
        array = array_in
    else:
        raise ValueError('axis_rows, axis_columns')

    shape_array = np.asarray(array.shape, dtype='i')
    n_axes = shape_array.size
    indexes_array = np.empty(n_axes, dtype='O')
    indexes_array[axis_rows] = slice(0, shape_array[axis_rows], 1)
    indexes_array[axis_columns] = slice(0, shape_array[axis_columns], 1)

    axes_array = np.arange(0, n_axes, 1)
    axes_table = np.asarray([axis_rows, axis_columns], dtype='i')
    axes_files = np.where(np.logical_not(cc_array.samples_in_arr1_are_in_arr2(axes_array, axes_table, axis=0)))[0]

    # for combination_indexes_i, combination_directories_i in cc_combinations.conditions_to_combinations_on_the_fly(
    #         conditions_of_directories, dtype='U', order_outputs='iv'):
    for combination_indexes_i, directory_i in cc_directory.conditions_to_directories_on_the_fly(
            conditions_of_directories, order_outputs='iv'):

        indexes_array[axes_files] = combination_indexes_i
        array_i = array[tuple(indexes_array)]

        dirname_i = os.path.dirname(directory_i)
        os.makedirs(dirname_i, exist_ok=True)
        with open(directory_i, 'w', newline='') as csv_file:
            writer = csv.writer(csv_file, delimiter=',')
            if headers is not None:
                writer.writerow(headers)
            writer.writerows(array_i)


def csv_file_to_lines(directory):
    with open(directory, newline='') as csvfile:
        reader = csv.reader(csvfile, delimiter=',')
        lines = list(reader)
    return lines


def csv_file_to_array(filename, rows, columns, dtype=None):

    with open(filename, newline='') as csvfile:
        reader = csv.reader(csvfile, delimiter=',')
        table = np.asarray(list(reader), dtype=dtype)

    indexes = tuple(cc_format.numeric_indexes_to_slice([rows, columns]))
    array = table[indexes]

    return array


def csv_file_to_arrays(filename, rows, columns, dtype=None):
    try:
        n_arrays_from_rows = len(rows)
    except TypeError:
        rows = [rows]
        n_arrays_from_rows = 1
    try:
        n_arrays_from_columns = len(columns)
    except TypeError:
        columns = [columns]
        n_arrays_from_columns = 1

    if n_arrays_from_rows != n_arrays_from_columns:
        if n_arrays_from_rows == 1:
            rows = list(rows) * n_arrays_from_columns
            n_arrays_from_rows = n_arrays_from_columns
        elif n_arrays_from_columns == 1:
            columns = list(columns) * n_arrays_from_rows
            n_arrays_from_columns = n_arrays_from_rows
        else:
            raise ValueError(
                'The following assumption is not met:\n'
                '\t n_arrays_from_rows' + ' \u003D ' + 'n_arrays_from_columns')

    n_arrays = n_arrays_from_rows
    try:
        n_dtypes = len(dtype)
    except TypeError:
        dtype = [dtype]
        n_dtypes = 1
    if n_arrays != n_dtypes:
        if n_dtypes == 1:
            dtype = list(dtype) * n_arrays
        else:
            raise ValueError(
                'The following assumption is not met:\n'
                '\t(n_dtypes \u003D n_arrays_from_rows) OR (n_dtypes \u003D n_arrays_from_columns)')

    with open(filename, newline='') as csvfile:
        reader = csv.reader(csvfile, delimiter=',')
        table = np.asarray(list(reader), dtype='U')

    n_axes_per_text = 2
    arrays = [None] * n_arrays
    for a in range(n_arrays):
        indexes = tuple(cc_format.numeric_indexes_to_slice([rows[a], columns[a]]))
        if (dtype[a] is None) or (dtype[a] == str) or ('U' in dtype[a][:3]):
            arrays[a] = table[indexes]
        else:
            arrays[a] = table[indexes].astype(dtype[a])
    return arrays


def conditions_of_csv_files_to_arrays(conditions_of_directories, rows, columns, dtype=None):

    n_axes_directories = len(conditions_of_directories)
    n_conditions_directories = np.empty(n_axes_directories, dtype=int)
    for i in range(n_axes_directories):
        n_conditions_directories[i] = len(conditions_of_directories[i])
    # logical_indexes_conditions = n_conditions_directories > 1
    combinations_directories = cc_combinations.n_conditions_to_combinations(n_conditions_directories)
    n_combinations_directories = combinations_directories.shape[0]
    axes_directories_squeezed = n_conditions_directories > 1
    n_axes_directories_squeezed = np.sum(axes_directories_squeezed)

    try:
        n_arrays_from_rows = len(rows)
    except TypeError:
        rows = [rows]
        n_arrays_from_rows = 1
    try:
        n_arrays_from_columns = len(columns)
    except TypeError:
        columns = [columns]
        n_arrays_from_columns = 1

    if n_arrays_from_rows != n_arrays_from_columns:
        if n_arrays_from_rows == 1:
            rows = list(rows) * n_arrays_from_columns
            n_arrays_from_rows = n_arrays_from_columns
        elif n_arrays_from_columns == 1:
            columns = list(columns) * n_arrays_from_rows
            n_arrays_from_columns = n_arrays_from_rows
        else:
            raise ValueError(
                'The following assumption is not met:\n'
                '\t n_arrays_from_rows' + ' \u003D ' + 'n_arrays_from_columns')

    n_arrays = n_arrays_from_rows
    try:
        n_dtypes = len(dtype)
    except TypeError:
        dtype = [dtype]
        n_dtypes = 1
    if n_arrays != n_dtypes:
        if n_dtypes == 1:
            dtype = list(dtype) * n_arrays
        else:
            raise ValueError(
                'The following assumption is not met:\n'
                '\t(n_dtypes \u003D n_arrays_from_rows) OR (n_dtypes \u003D n_arrays_from_columns)')

    n_axes_per_csv = 2
    indexes = [None] * n_arrays  # type: list
    arrays = [None] * n_arrays  # type: list
    for a in range(n_arrays):
        indexes[a] = tuple(cc_format.numeric_indexes_to_slice([rows[a], columns[a]]))

    directory_csv_d = os.path.join(*[
        conditions_of_directories[i][0] for i in range(n_axes_directories)])

    with open(directory_csv_d, newline='') as csvfile:
        reader = csv.reader(csvfile, delimiter=',')
        # array_csv_d = list(reader)
        array_csv_d = np.asarray(list(reader), dtype='U')

    indexes_out = [None] * n_arrays
    for a in range(n_arrays):
        if dtype[a] is None:
            dtype[a] = 'U'

        array_a_d = array_csv_d[indexes[a]]
        shape_array_a_d = np.asarray(array_a_d.shape, dtype=int)
        shape_array_a = np.append(
            n_conditions_directories[axes_directories_squeezed], shape_array_a_d)

        arrays[a] = np.empty(shape_array_a, dtype=dtype[a])
        n_axes_a = shape_array_a.size
        indexes_out = np.empty(n_axes_a, dtype=object)
        indexes_out[:] = slice(None)

    for d in range(n_combinations_directories):

        directory_csv_d = os.path.join(*[
            conditions_of_directories[i][combinations_directories[d, i]] for i in range(n_axes_directories)])

        with open(directory_csv_d, newline='') as csvfile:
            reader = csv.reader(csvfile, delimiter=',')
            array_csv_d = np.asarray(list(reader), dtype='U')

        indexes_out[slice(0, n_axes_directories_squeezed, 1)] = (
            combinations_directories[d, axes_directories_squeezed])

        for a in range(n_arrays):
            arrays[a][tuple(indexes_out)] = array_csv_d[indexes[a]]  # .astype(dtype[a])

    return arrays


def n_csv_files_to_array_old(
        directories_csv_files, names_csv_files_in_directories=False,
        rows=slice(None), columns=slice(None), data_type=None):

    print('using the funtion n_csv_files_to_array().\n'
          'In the future versions of ccalafiore, it will be removed.\n'
          'Consider using conditions_of_csv_files_to_arrays()')
    array = None

    n_axes_directories = len(directories_csv_files)
    n_conditions_directories = np.empty(n_axes_directories, dtype=int)
    for a in range(n_axes_directories):
        n_conditions_directories[a] = len(directories_csv_files[a])

    logical_indexes_conditions = n_conditions_directories > 1

    combinations_directories = cc_combinations.n_conditions_to_combinations(n_conditions_directories)
    n_combinations_directories = combinations_directories.shape[0]

    for d in range(n_combinations_directories):

        directory = directories_csv_files[0][combinations_directories[d, 0]]
        for a in range(1, n_axes_directories):
            directory = os.path.join(
                directory, directories_csv_files[a][combinations_directories[d, a]])

        if names_csv_files_in_directories:
            files_per_directory = [directory]
        else:
            files_per_directory = glob.glob(os.path.join(directory, '*.csv'))

        n_files_per_directory = len(files_per_directory)

        start_row_f_file = 0
        for f in range(n_files_per_directory):

            array_per_file = np.genfromtxt(files_per_directory[f], delimiter=',', dtype=data_type)[rows, columns]

            if array is None:

                shape_array_per_file = np.asarray(array_per_file.shape, dtype=int)

                shape_array = np.asarray([
                    *n_conditions_directories[logical_indexes_conditions],
                    n_files_per_directory * shape_array_per_file[0],
                    *shape_array_per_file[1:]
                ], dtype=int)

                array = np.empty(shape_array, dtype=data_type)

            end_row_f_file = (f + 1) * shape_array_per_file[0]

            indexes_array = tuple(
                [*combinations_directories[d][logical_indexes_conditions],
                 slice(start_row_f_file, end_row_f_file)])

            array[indexes_array] = array_per_file
            start_row_f_file = end_row_f_file

    return array



import pickle


def save(data, directory_file):
    outfile = open(directory_file, 'wb')
    pickle.dump(data, outfile)
    outfile.close()
    return None


def load(directory_file):
    outfile = open(directory_file, "rb")
    data = pickle.load(outfile)
    outfile.close()
    return data




import numpy as np


def numeric_indexes_to_slice(indexes):

    if not isinstance(indexes, slice):
        try:
            n_indexes_a = len(indexes)
            if n_indexes_a == 0:
                pass
            elif n_indexes_a == 1:
                # indexes = slice(indexes[0], indexes[0] + 1, 1)
                indexes = indexes[0]
            elif n_indexes_a > 1:
                sliceable = True
                step = indexes[1] - indexes[0]
                for i in range(2, n_indexes_a):
                    step_i = indexes[i] - indexes[i - 1]
                    if step_i != step:
                        sliceable = False
                        break
                if sliceable:
                    if step > 0:
                        stop = indexes[-1] + 1
                    else:
                        stop = indexes[-1] - 1
                        if stop < 0:
                            stop = None
                    indexes = slice(indexes[0], stop, step)

        except TypeError:
            pass

    return indexes


def seq_of_numeric_indexes_to_seq_of_slices(indexes):
    n_axes = len(indexes)
    for a in range(n_axes):
        indexes[a] = numeric_indexes_to_slice(indexes[a])
    return indexes


def format_shape_arguments(dict_arguments, target_shape):

    if isinstance(target_shape, list) or isinstance(target_shape, tuple):
        target_shape = np.asarray(target_shape, dtype=int)
    if isinstance(target_shape, np.ndarray):
        pass
    else:
        target_shape = np.asarray([target_shape], dtype=int)

    n_axes = target_shape.size

    # n_arguments = len(arguments)
    for key_i, argument_i in dict_arguments.items():

        # format arguments[i]
        if argument_i is None:
            argument_i = np.empty(target_shape, dtype=object)

        elif isinstance(argument_i, str):

            len_argument_i = len(argument_i)
            dtype_argument_i = '<U' + str(len_argument_i)
            argument_i_tmp = argument_i
            argument_i = np.empty(target_shape, dtype=dtype_argument_i)
            argument_i[:] = argument_i_tmp

        elif (isinstance(argument_i, list) or
              isinstance(argument_i, np.ndarray) or
              isinstance(argument_i, tuple)):

            if isinstance(argument_i, list) or isinstance(argument_i, tuple):
                # argument_i = np.asarray(argument_i)
                argument_i = np.asarray(argument_i)

            shape_argument_i = np.asarray(argument_i.shape, dtype=int)
            shape_in_argument_i = shape_argument_i[slice(0, n_axes, 1)]
            n_axis_in_argument_i = shape_in_argument_i.size
            if n_axis_in_argument_i != n_axes:
                shape_new_argument_i = np.append(target_shape, shape_argument_i)
                n_axis_in_new_argument_i = shape_new_argument_i.size
                dtype = argument_i.dtype
                argument_i_tmp = argument_i
                for a in range(n_axes):
                    argument_i_tmp = np.expand_dims(argument_i_tmp, axis=0)
                argument_i = np.empty(shape_new_argument_i, dtype=dtype)
                index = np.empty(n_axis_in_new_argument_i, dtype=object)
                index[:] = slice(None)
                argument_i[tuple(index)] = argument_i_tmp

            elif (shape_in_argument_i == target_shape).all():
                pass
            elif (shape_in_argument_i[shape_in_argument_i != target_shape] == 1).all():
                shape_argument_i = shape_argument_i[slice(n_axes, None, 1)]
                shape_new_argument_i = np.append(target_shape, shape_argument_i)
                n_axis_in_new_argument_i = shape_new_argument_i.size
                dtype = argument_i.dtype
                argument_i_tmp = argument_i
                argument_i = np.empty(shape_new_argument_i, dtype=dtype)
                index = np.empty(n_axis_in_new_argument_i, dtype=object)
                index[:] = slice(None)
                argument_i[tuple(index)] = argument_i_tmp[tuple(index)]
            else:
                shape_new_argument_i = np.append(target_shape, shape_argument_i)
                n_axis_in_new_argument_i = shape_new_argument_i.size
                dtype = argument_i.dtype
                argument_i_tmp = argument_i
                for a in range(n_axes):
                    argument_i_tmp = np.expand_dims(argument_i_tmp, axis=0)
                argument_i = np.empty(shape_new_argument_i, dtype=dtype)
                index = np.empty(n_axis_in_new_argument_i, dtype=object)
                index[:] = slice(None)
                argument_i[tuple(index)] = argument_i_tmp
        else:
            dtype_argument_i = type(argument_i)
            argument_i_tmp = argument_i
            argument_i = np.empty(target_shape, dtype=dtype_argument_i)
            argument_i[:] = argument_i_tmp

        dict_arguments[key_i] = argument_i

    return dict_arguments


def dict_to_key_array_and_value_array(dictionary):
    n_items = len(dictionary)
    keys = [None] * n_items
    values = np.empty(n_items, dtype=int)

    max_len_keys = 0
    i = 0
    for k, v in dictionary.items():
        keys[i] = k
        len_k = len(k)
        if len_k > max_len_keys:
            max_len_keys = len_k

        values[i] = v

        i += 1

    keys = np.asarray(keys, dtype=('<U' + str(max_len_keys)))
    return keys, values




from os.path import isfile as os_path_isfile
from time import sleep
from . import clock as cc_clock


def wait_downloading(directory_saved_as, max_seconds_wait=60):

    timer = cc_clock.Timer()
    wait = True
    while wait:

        if os_path_isfile(directory_saved_as):
            wait = False
        elif timer.get_seconds() > max_seconds_wait:
            raise TimeoutError
        sleep(1)






import numpy as np
import math
import numbers
# from copy import deepcopy
# from . import combinations as cc_combinations
from . import maths as cc_maths
# from .maths import convert_to_int_or_float, rint


def advanced_indexing(indexes_loose):
    def clean_input_of_ix_(indexes_loose):
        n_axes = len(indexes_loose)
        for a in range(n_axes):
            try:
                len(indexes_loose[a])
            except TypeError:
                indexes_loose[a] = np.asarray([indexes_loose[a]], dtype=int)
        return indexes_loose

    indexes_loose = clean_input_of_ix_(indexes_loose)
    indexes = np.ix_(*indexes_loose)
    return indexes


def samples_in_arr1_are_not_in_arr2(arr1, arr2, axis=0):
    return np.logical_not(samples_in_arr1_are_in_arr2(arr1, arr2, axis=axis))


def samples_in_arr1_are_in_arr2(arr1, arr2, axis=0):

    # OLD elements_in_x_are_in_y()

    arr1, arr2 = make_sure_is_iterable(arr1, arr2)

    arr1 = np.array(arr1)
    type_arr1 = arr1.dtype
    if type_arr1 == object:
        arr2 = np.array(list(arr1))

    arr2 = np.array(arr2)
    type_arr2 = arr2.dtype
    if type_arr2 == object:
        arr2 = np.array(list(arr2))

    shape_arr1 = np.array(arr1.shape)
    shape_arr2 = np.array(arr2.shape)

    n_axes_arr1 = len(shape_arr1)
    n_axes_arr2 = len(shape_arr2)
    if n_axes_arr1 != n_axes_arr2:

        if n_axes_arr1 == n_axes_arr2 - 1:

            arr1 = np.expand_dims(arr1, axis=axis)
            shape_arr1 = np.array(arr1.shape)
            n_axes_arr1 = len(shape_arr1)

        else:
            raise Exception('n_axes_arr1 has to be equal to n_axes_arr2\n'
                            'or to n_axes_arr2 - 1.\n'
                            'Now, n_axes_arr1 = {} and n_axes_arr2 = {}.'.format(
                n_axes_arr1, n_axes_arr2))

    axes_arr1_and_arr2 = np.arange(n_axes_arr1)

    if np.all(axis != axes_arr1_and_arr2):
        raise Exception('axis_1 is not in the axes_arr1_and_arr2.\n'
                        'axis_1 has to be equal to either one of the\n'
                        'axes_arr1_and_arr2.\n'
                        'Now, axis_1 = {} and axes_arr1_and_arr2 = {}.\n'
                        'By default, axis_1=0.'.format(
                         axis, axes_arr1_and_arr2))

    n_samples_arr1 = shape_arr1[axis]
    n_samples_arr2 = shape_arr2[axis]

    axes_within_samples = np.argwhere(axes_arr1_and_arr2 != axis)[:, 0]
    # dimensions_within_samples = dimensions_within_samples.astype(int)

    shape_1_sample_arr1 = shape_arr1[axes_within_samples]
    shape_1_sample_arr2 = shape_arr2[axes_within_samples]
    if np.any(shape_1_sample_arr1 != shape_1_sample_arr2):

        raise Exception('shape_1_sample_arr1 has to be equal to shape_1_sample_arr2.\n'
                        'Now,shape_1_sample_arr1 = {} and shape_1_samples_arr2 = {}.'.format(
            shape_1_sample_arr1, shape_1_sample_arr2))

    shape_arr_logical = np.copy(shape_arr2)
    shape_arr_logical = np.insert(shape_arr_logical, axis, shape_arr1[axis], axis=0)
    n_axes_arr_logical = len(shape_arr_logical)
    arr_logical = np.empty(shape_arr_logical, dtype=bool)

    indexes_arr1 = np.full(n_axes_arr1, slice(None))
    indexes_arr_logical = np.full(n_axes_arr_logical, slice(None))

    # indexes_arr1 = np.empty(n_dimensions_arr1, dtype=object)
    # indexes_arr_logical = np.empty(n_dimensions_arr_logical, dtype=object)
    # for i_dimension in range(n_dimensions_arr_logical):
    #     if i_dimension != axis_1:
    #         # indexes_samples_i_window.append(slice(0, n_samples_window))
    #     #     indexes_arr1[i_dimension] = np.arange([0])
    #     # else:
    #         indexes_arr_logical[i_dimension] = np.arange(shape_arr_logical[i_dimension])
    #         if i_dimension < n_dimensions_arr1:
    #             indexes_arr1[i_dimension] = np.arange(shape_arr1[i_dimension])

    for i_sample in range(n_samples_arr1):
        indexes_arr1[axis] = i_sample
        indexes_arr_logical[axis] = i_sample
        arr_logical[tuple(indexes_arr_logical)] = arr1[tuple(indexes_arr1)] == arr2

        # maybe todo: arr_logical[tuple(indexes_arr_logical)] = np.all(arr1[tuple(indexes_arr1)] == arr2)

    elements_of_arr1_in_arr2_logical = np.any(arr_logical, axis=axis + 1)

    return elements_of_arr1_in_arr2_logical


def make_sure_is_iterable(*variables, if_not_iterable_convert_2_type_variable=np.ndarray):
    # list_types_iterable_allowed = [list, tuple, np.ndarray, str]

    # n_types_iterable_allowed = len(list_types_iterable_allowed)
    # if n_types_iterable_allowed == 1:
    #     if_not_iterable_convert_2_type_variable = list_types_iterable_allowed[0]
    # else:
    #     if if_not_iterable_convert_2_type_variable not in list_types_iterable_allowed:
    #         raise Exception(
    #             'if_not_iterable_convert_2_type_variable is not in list_types_iterable_allowed.\n'
    #             'Now, if_not_iterable_convert_2_type_variable = {} and list_types_iterable_allowed = {}.\n'
    #             'By default, make_sure_is_iterable(*variables, list_types_iterable_allowed=[list, tuple, np.ndarray],\n'
    #             'if_not_iterable_convert_2_type_variable=np.ndarray)'.format(
    #                 if_not_iterable_convert_2_type_variable, list_types_iterable_allowed))

    variables = list(variables)
    n_variables = len(variables)

    for i in range(n_variables):

        try:
            len(variables[i])

        except TypeError:

            if if_not_iterable_convert_2_type_variable == str:
                variables[i] = str(variables[i])

            else:
                variables[i] = [variables[i]]

                if if_not_iterable_convert_2_type_variable == list:
                    variables[i] = list(variables[i])

                elif if_not_iterable_convert_2_type_variable == tuple:
                    variables[i] = tuple(variables[i])

                elif if_not_iterable_convert_2_type_variable == np.ndarray:
                    variables[i] = np.array(variables[i])
                else:
                    raise Exception(
                        'if_not_iterable_convert_2_type_variable has to either be equal to np.ndarray, list, tuple or str.\n'
                        'Now, if_not_iterable_convert_2_type_variable = {}.\n'
                        'By default, if_not_iterable_convert_2_type_variable=np.ndarray'.format(
                            if_not_iterable_convert_2_type_variable))

        # type_i_variable = type(variables[i])
        #
        # # if type_i_variable not in list_types_iterable_allowed:
        #     if if_not_iterable_convert_2_type_variable == list:
        #         variables[i] = list([variables[i]])
        #
        #     elif if_not_iterable_convert_2_type_variable == tuple:
        #         variables[i] = tuple([variables[i]])
        #
        #     elif if_not_iterable_convert_2_type_variable == np.ndarray:
        #         variables[i] = np.array([variables[i]])
        #     else:
        #         raise Exception(
        #             'if_not_iterable_convert_2_type_variable has to either be equal to np.ndarray, list or tuple.\n'
        #             'Now, if_not_iterable_convert_2_type_variable = {}.\n'
        #             'By default, if_not_iterable_convert_2_type_variable=np.ndarray'.format(
        #                 if_not_iterable_convert_2_type_variable))

        # elif type_i_variable == np.ndarray:
        #     if variables[i].ndim == 0:
        #         variables[i] = np.array([variables[i]])

    if n_variables == 1:
        return variables[0]
    else:
        return variables


def transfer_n_random_samples_from_arr1_to_arr2(arr1, arr2, n_samples=1, axis=0, replace=False):
    # def transfer_n_random_samples_from_arr1_to_arr2(arr1, arr2, first_n_samples=1, all=False):

    # SPLIT THIS IN TWO FUNCTIONS:
    # 1) transfer_n_random_samples_from_arr1_to_arr2;
    # 2) transfer_n_samples_from_arr1_to_arr2;

    arr1, arr2 = make_sure_is_iterable(arr1, arr2)


    shape_arr1 = np.array(arr1.shape)
    shape_arr2 = np.array(arr2.shape)

    n_dimensions_arr1 = len(shape_arr1)
    n_dimensions_arr2 = len(shape_arr2)
    if n_dimensions_arr1 != n_dimensions_arr2:

        if n_dimensions_arr1 == n_dimensions_arr2 - 1:

            arr1 = np.expand_dims(arr1, axis=axis)
            shape_arr1 = np.array(arr1.shape)
            n_dimensions_arr1 = len(shape_arr1)

        else:
            raise Exception('n_dimensions_arr1 has to be equal to n_dimensions_arr2\n'
                            'or to n_dimensions_arr2 - 1.\n'
                            'Now, n_dimensions_arr1 = {} and n_dimensions_arr2 = {}.'.format(
                n_dimensions_arr1, n_dimensions_arr2))

    dimensions_arr1_and_arr2 = np.arange(n_dimensions_arr1)

    if np.all(axis != dimensions_arr1_and_arr2):
        raise Exception('axis is not in the dimensions_arr1_and_arr2.\n'
                        'axis has to be equal to either one ot the\n'
                        'dimensions_arr1_and_arr2.\n'
                        'Now, axis = {} and dimensions_arr1_and_arr2 = {}.\n'
                        'By default, axis=0.'.format(
            axis, dimensions_arr1_and_arr2))

    n_samples_arr1 = shape_arr1[axis]
    n_samples_arr2 = shape_arr2[axis]
    if n_samples > n_samples_arr1 and replace is False:
        raise Exception('n_samples cannot be greater than n_samples of arr1, when replace=False.\n'
                        'Now, n_samples = {}, n_samples_arr1 = {} and replace = {}.'.format(
            n_samples, n_samples_arr1, replace))


    dimensions_within_samples = np.argwhere(dimensions_arr1_and_arr2 != axis)[:, 0]
    # dimensions_within_samples = dimensions_within_samples.astype(int)

    shape_1_sample_arr1 = shape_arr1[dimensions_within_samples]
    shape_1_sample_arr2 = shape_arr2[dimensions_within_samples]
    if np.any(shape_1_sample_arr1 != shape_1_sample_arr2):

        raise Exception('shape_1_sample_arr1 has to be equal to shape_1_sample_arr2.\n'
                        'Now,shape_1_sample_arr1 = {} and shape_1_samples_arr2 = {}.'.format(
            shape_1_sample_arr1, shape_1_sample_arr2))

    indexes_chosen_samples_in_arr1 = np.random.choice(n_samples_arr1, n_samples, replace=replace)

    indexes = np.empty(n_dimensions_arr1, dtype=object)
    for i_d in range(n_dimensions_arr1):
        if i_d == axis:
            # indexes_samples_i_window.append(slice(0, n_samples_window))
            indexes[i_d] = indexes_chosen_samples_in_arr1
        else:
            indexes[i_d] = np.arange(shape_arr1[i_d])

    # if n_dimensions_arr1 == 1:
    #     chosen_sample_in_arr1 = arr1[np.ix_(indexes)]
    # else:
    chosen_sample_in_arr1 = arr1[np.ix_(*indexes)]

    arr2 = np.append(arr2, chosen_sample_in_arr1)

    arr1 = np.delete(arr1, indexes_chosen_samples_in_arr1, axis=axis)
    # indexes_chosen_sample_in_arr1 = np.argwhere(arr1 == chosen_sample_in_arr1)[0]

    # n_chosen_samples_in_arr1 = len(indexes_chosen_sample_in_arr1)
    # if all:
    #
    # elif n_chosen_samples_in_arr1 <= first_n_samples:
    #     indexes_chosen_sample_in_arr1 = indexes_chosen_sample_in_arr1[np.arange(n_chosen_samples_in_arr1)]
    #
    # elif n_chosen_samples_in_arr1 > first_n_samples:
    #     indexes_chosen_sample_in_arr1 = indexes_chosen_sample_in_arr1[np.arange(first_n_samples)]

    # arr1 = np.delete(arr1, indexes_chosen_sample_in_arr1)

    return arr1, arr2


def shuffle_in_windows(data, n_samples_window=None, n_windows=None, axis=0, dtype=None):

    type_data = type(data)
    if type_data != np.ndarray:
        raise Exception('data must be an numpy array. '
                        'Now, the class of data is {}'.format(type_data))

    shape_data = data.shape
    n_samples_data = shape_data[axis]
    n_d_data = len(shape_data)
    if axis > n_d_data:
        raise Exception(
            'shuffle in dimension {} can\'t be done,\n'
            'because data has {} dimensions and\n'
            'it doesn\'t have the {}th dimension.\n'
            'axis has to be an integer in the interval\n'
            '0 =< axis < n_d_data.\n'
            'In this case, axis = {} and n_d_data = {}'.format(
                axis, n_d_data, axis, axis, n_d_data))

    if n_samples_window is None and n_windows is None:
        raise Exception('Both n_samples_window and n_windows are None.'
                        'Define only one of them with an integer.'
                        'One of them has to be None and the other one has to be an integer')

    elif n_samples_window is None and n_windows is not None:
        n_samples_window = math.ceil(n_samples_data / n_windows)

    elif n_samples_window is not None and n_windows is None:
        n_windows = math.ceil(n_samples_data / n_samples_window)

    elif n_samples_window is not None and n_windows is not None:

        tmp_n_samples_window = math.ceil(n_samples_data / n_windows)
        tmp_n_windows = math.ceil(n_samples_data / n_samples_window)

        if tmp_n_samples_window != n_samples_window and tmp_n_windows != n_windows:

            raise Exception(
                'For n_samples_data = {}, n_samples_window = {} and n_windows = {} are not compatible.\n'
                'If n_samples_window = {}, n_windows has to be {} or None.\n'
                'If n_windows = {}, n_samples_window has to be {} or None'.format(
                    n_samples_data, n_samples_window, n_windows, n_samples_window,
                    tmp_n_windows, n_windows, tmp_n_samples_window))

    if n_samples_window * n_windows > n_samples_data:
        print('Warning: n_samples_window * n_windows > n_samples_data.\n'
              'n_samples_window * n_windows = {} and n_samples_data = {}'.format(
            n_samples_window * n_windows, n_samples_data))

    indexes_i_window = np.empty(n_d_data, dtype=object)
    for i_d in range(n_d_data):
        if i_d != axis:
            indexes_i_window[i_d] = slice(0, shape_data[i_d], 1)

    if dtype is None:
        dtype = data.dtype
    data_shuffle = np.empty(shape_data, dtype=dtype)
    start = 0
    end = n_samples_window
    for i_window in range(n_windows):

        indexes_i_window[axis] = slice(start, end, 1)
        indexes_tuple = tuple(indexes_i_window)
        data_shuffle[indexes_tuple] = shuffle_in_any_dimension(data[indexes_tuple], axis=axis)
        start = end
        end += n_samples_window

    return data_shuffle


def shuffle_in_any_dimension(data, axis=0):

    type_data = type(data)
    if type_data != np.ndarray:
        raise Exception('data must be an numpy array. '
                        'Now, the class of data is {}'.format(type_data))

    shape_data = data.shape
    n_samples = shape_data[axis]
    n_axes = len(shape_data)

    if axis > n_axes:
        raise Exception(
            'shuffle in axis {} can\'t be done,\n'
            'because data has {} axes and\n'
            'it doesn\'t have the {}th axis.\n'
            'axis has to be an integer in the interval\n'
            '0 =< axis < n_axes.\n'
            'In this case, axis = {} and n_axes = {}'.format(
                axis, n_axes, axis, axis, n_axes))


    indexes_new = np.empty(n_axes, dtype=object)
    for a in range(n_axes):
        if a == axis:
            # indexes_samples_i_window.append(slice(0, n_samples_window))
            indexes_new[a] = np.random.rand(n_samples).argsort()
        else:
            indexes_new[a] = slice(0, shape_data[a], 1)

    data_shuffle = data[tuple(indexes_new)]

    return data_shuffle


def split_by_n(array, n_slices, axis=0):

    shape_array = np.array(array.shape)

    n_dimensions_array = len(shape_array)

    indexes = np.empty(n_dimensions_array, dtype=object)

    for i_d in range(n_dimensions_array):
        if i_d != axis:

            indexes[i_d] = slice(None)

    # indexes_samples_i_window.append(slice(0, n_samples_window))

    n_elements_array = shape_array[axis]

    n_elements_slice = n_elements_array / n_slices

    sliced_array = np.empty(n_slices, dtype=object)

    index_start = 0
    for i_slice in range(n_slices):

        index_stop = int((i_slice + 1) * n_elements_slice)

        indexes[axis] = slice(index_start, index_stop)

        # sliced_array[i_slice] = array[int(i_slice * n_elements_slice): int((i_slice + 1) * n_elements_slice)]
        sliced_array[i_slice] = array[tuple(indexes)]

        index_start = index_stop

    return sliced_array


def split_by_percentages(array, p_samples, axis=0, shuffle=False, sort=False):

    N = array.shape[axis]
    S = len(p_samples)
    M = np.empty(S, dtype='i')

    for s in range(S):
        M[s] = cc_maths.rint(N / 100 * p_samples[s])

    sum_M = M.sum()
    if sum_M != N:
        if sum_M < N:
            s = 0
            M[s] += 1
            while sum_M < N:
                s += 1
                M[s] += 1
        else:
            s = 0
            M[s] -= 1
            while sum_M < N:
                s += 1
                M[s] -= 1


    indexes_splitting = np.empty(S - 1, dtype='i')
    for s in range(S - 1):
        indexes_splitting[s] = M[slice(0, s + 1, 1)].sum()

    if shuffle:
        n_axes = len(array.shape)
        indexes_array = np.empty(n_axes, dtype=object)
        indexes_array[:] = slice(0, None, 1)
        indexes_array[axis] = np.random.permutation(N)
        arrays = np.split(array[tuple(indexes_array)], indexes_splitting, axis=axis)

        # arrays = np.split(shuffle_in_any_dimension(array, axis=axis), indexes_splitting, axis=axis)
        if sort:
            indexes_perm = np.split(indexes_array[axis], indexes_splitting, axis=axis)
            indexes_recover = [None] * S  # type: list
            # arrays_new = [None] * S  # type: list
            for s in range(S):
                indexes_recover[s] = np.empty(M[s], dtype='i')
                start_s = 0
                for m in range(M[s]):
                    for i in range(start_s, N):
                        indexes_found = np.where(i == indexes_perm[s])[0]
                        n_indexes_found = len(indexes_found)
                        if n_indexes_found == 1:
                            indexes_recover[s][m] = indexes_found[0]
                            # np.delete(indexes_perm[s], indexes_found)
                            start_s = i + 1
                            break

                indexes_array[axis] = indexes_recover[s]
                arrays[s] = arrays[s][tuple(indexes_array)]
                # arrays_new[s] = arrays[s][tuple(indexes_array)]
                # indexes_recover[s] = [np.where(i == indexes_perm[s])[0][0] for i in range(N)]
                # c = b[indexes_recover]
                # arrays[s].sort()
    else:
        arrays = np.split(array, indexes_splitting, axis=axis)

    return arrays


def pad_array_from_n_samples_target(array, n_samples_target=1, axis=0):

    try:
        len(array)
        array = np.asarray(array)
    except TypeError:
        array = np.asarray([array])

    shape_array = np.asarray(array.shape)
    n_axes = len(shape_array)
    while n_axes <= axis:
        array = np.expand_dims(array, axis=n_axes)
        shape_array = np.asarray(array.shape)
        n_axes = len(shape_array)

    n_samples_true = shape_array[axis]
    while n_samples_true < n_samples_target:
        array = np.append(array, array, axis=axis)
        n_samples_true = array.shape[axis]

    index_array = np.full(n_axes, slice(None))
    index_array[axis] = slice(n_samples_target)
    array = array[tuple(index_array)]

    return array


def pad_arr_1_from_arr_2(arr_1, arr_2, axis=0):

    try:
        len(arr_1)
        arr_1 = np.asarray(arr_1)
    except TypeError:
        arr_1 = np.asarray([arr_1])

    try:
        len(arr_2)
        arr_2 = np.asarray(arr_2)
    except TypeError:
        arr_2 = np.asarray([arr_2])

    shape_arr_1 = np.asarray(arr_1.shape)
    n_axes_1 = len(shape_arr_1)
    shape_arr_2 = np.asarray(arr_2.shape)
    n_axes_2 = len(shape_arr_2)

    if n_axes_1 == n_axes_2 - 1:
        arr_1 = np.expand_dims(arr_1, axis=axis)
        shape_arr_1 = np.asarray(arr_1.shape)
        n_axes_1 = len(shape_arr_1)
    elif n_axes_1 == n_axes_2:
        pass
    else:
        raise Exception('dimensions of arr_1 and arr_2 do not match')

    if np.all(shape_arr_1[np.arange(n_axes_1) != axis] == shape_arr_2[np.arange(n_axes_2) != axis]):
        pass
    else:
        raise Exception('dimensions of arr_1 and arr_2 do not match')

    n_samples_true = shape_arr_1[axis]
    n_samples_target = shape_arr_2[axis]

    while n_samples_true < n_samples_target:
        arr_1 = np.append(arr_1, arr_1, axis=axis)
        n_samples_true = arr_1.shape[axis]

    index_arr_1 = np.full(n_axes_1, slice(None))
    index_arr_1[axis] = slice(n_samples_target)

    arr_1 = arr_1[tuple(index_arr_1)]

    return arr_1


def centered_indexes_to_normal_indexes(
        centered_indexes, axis_indexes, shape_array):

    axis_dimensions_of_array = int(not (axis_indexes == 1))

    if isinstance(centered_indexes, list) or isinstance(centered_indexes, tuple):
        centered_indexes = np.asarray(centered_indexes)
        n_axes_indexes = len(centered_indexes.shape)
    elif isinstance(centered_indexes, np.ndarray):
        n_axes_indexes = len(centered_indexes.shape)
    else:
        n_axes_indexes = 0

    if isinstance(shape_array, list) or isinstance(shape_array, tuple):
        shape_array = np.asarray(shape_array)
        n_axes_shape_array = shape_array.size
    elif isinstance(shape_array, np.ndarray):
        n_axes_shape_array = shape_array.size
    else:
        n_axes_shape_array = 0

    if (n_axes_indexes == 2) and (n_axes_shape_array == 1):
        shape_array = np.expand_dims(shape_array, axis=axis_indexes)

    normal_indexes = (shape_array / 2) + centered_indexes
    if isinstance(normal_indexes, np.ndarray):
        normal_indexes = np.floor(normal_indexes)
    else:
        normal_indexes = math.floor(normal_indexes)

    return normal_indexes


class IntRange:

    def __init__(self, raw_range):

        if isinstance(raw_range, IntRange):
            start = raw_range.start
            stop = raw_range.stop
            step = raw_range.step

        elif isinstance(raw_range, dict):
            keys = ['start', 'stop', 'step']
            for k in raw_range.keys():
                if k in keys:
                    continue
                else:
                    raise KeyError(k)
            if raw_range.get('start') is None:
                start = 0
            else:
                start = raw_range['start']

            if raw_range.get('stop') is None:
                raise ValueError("raw_range['stop']")
            else:
                stop = raw_range['stop']

            if raw_range.get('step') is None:
                step = 1
            else:
                step = raw_range['step']

        elif isinstance(raw_range, range):
            start = raw_range.start
            stop = raw_range.stop
            step = raw_range.step
        elif isinstance(raw_range, (list, tuple, np.ndarray)):
            n_elements = len(raw_range)
            if n_elements == 1:
                start = 0
                stop = raw_range[0]
                step = 1
            elif n_elements == 2:
                start, stop = raw_range
                step = 1
            elif n_elements == 3:
                start, stop, step = raw_range
            else:
                raise ValueError('raw_range has to have at least 1 element. {} are given'.format(n_elements))
        elif isinstance(raw_range, numbers.Number):
            start = 0
            stop = raw_range
            step = 1
        else:
            raise TypeError('raw_range')

        self.start = start
        self.stop = stop
        self.step = step
        if self.stop > self.start:
            if self.step > 0:
                self.len = math.ceil((self.stop - self.start) / self.step)
            elif self.step < 0:
                self.len = 0
            else:
                raise ValueError('self.step')
        elif self.stop < self.start:
            if self.step < 0:
                self.len = math.ceil(abs((self.stop - self.start) / self.step))
            elif self.step > 0:
                self.len = 0
            else:
                raise ValueError('self.step')
        else:
            self.len = 0

    def to_slice(self):
        return slice(self.start, self.stop, self.step)

    def to_list(self):
        return list(range(self.start, self.stop, self.step))


class FloatRange:
    def __init__(self, raw_range):
        if isinstance(raw_range, dict):
            keys = ['start', 'stop', 'step']
            for k in raw_range.keys():
                if k in keys:
                    continue
                else:
                    raise KeyError(k)
            if raw_range.get('start') is None:
                start = 0
            else:
                start = raw_range['start']

            if raw_range.get('stop') is None:
                raise ValueError("raw_range['stop']")
            else:
                stop = raw_range['stop']

            if raw_range.get('step') is None:
                step = 1
            else:
                step = raw_range['step']

        elif isinstance(raw_range, range):
            start = raw_range.start
            stop = raw_range.stop
            step = raw_range.step
        elif isinstance(raw_range, (list, tuple, np.ndarray)):
            n_elements = len(raw_range)
            if n_elements == 1:
                start = 0
                stop = raw_range[0]
                step = 1
            elif n_elements == 2:
                start, stop = raw_range
                step = 1
            elif n_elements == 3:
                start, stop, step = raw_range
            else:
                raise ValueError('raw_range has to have at least 1 element. {} are given'.format(n_elements))
        elif isinstance(raw_range, numbers.Number) or raw_range is None:
            start = 0
            stop = raw_range
            step = 1
        else:
            raise TypeError('raw_range')

        self.start = cc_maths.convert_to_int_or_float(start)
        self.stop = cc_maths.convert_to_int_or_float(stop)
        self.step = cc_maths.convert_to_int_or_float(step)
        self.len = abs(cc_maths.convert_to_int_or_float((self.stop - self.start) / self.step))

    # def __init__(self, stop, *args):
    #
    #     n_args = len(args)
    #     if n_args == 0:
    #         self.start = 0
    #         self.stop = stop
    #         self.step = 1
    #     elif n_args == 1:
    #         self.start = stop
    #         self.stop = args[0]
    #         self.step = 1
    #     elif n_args == 2:
    #         self.start = stop
    #         self.stop = args[0]
    #         self.step = args[1]
    #     else:
    #         raise TypeError('MyRange expected at most 3 arguments, got {}'.format(n_args + 1))
    #
    #     self.len = cc_maths.convert_to_int_or_float((self.stop - self.start) / self.step)
    #     # self.len = (self.stop - self.start) / self.step

    def starts_of_n_within_equal_ranges(self, n):

        len_each_equal_ranges = self.len / n
        starts = np.arange(
            start=self.start,
            stop=self.stop - (len_each_equal_ranges * .1),
            step=len_each_equal_ranges)
        return starts

    def stops_of_n_within_equal_ranges(self, n):

        len_each_equal_ranges = self.len / n
        stops = np.arange(
            start=self.start + len_each_equal_ranges,
            stop=self.stop + len_each_equal_ranges - (len_each_equal_ranges * .1),
            step=len_each_equal_ranges)

        return stops

    def centers_of_n_within_equal_ranges(self, n):

        len_each_equal_ranges = self.len / n
        half_len_each_equal_ranges = len_each_equal_ranges / 2
        centers = np.arange(
            start=self.start + half_len_each_equal_ranges,
            stop=self.stop + half_len_each_equal_ranges - (len_each_equal_ranges * .1),
            step=len_each_equal_ranges)
        return centers


def select_dtype_from_2_arrays(x, y):

    if isinstance(x, np.ndarray):
        dtype_x = x.dtype
    else:
        dtype_x = np.asarray(x).dtype

    if isinstance(y, np.ndarray):
        dtype_y = y.dtype
    else:
        dtype_y = np.asarray(y).dtype

    if dtype_x.kind == dtype_y.kind:
        if int(dtype_x.str[2:]) > int(dtype_y.str[2:]):
            dtype_out = dtype_x.str
        else:
            dtype_out = dtype_y.str

    elif (dtype_x.kind == 'f') and (dtype_y.kind == 'i'):
        dtype_out = dtype_x.str
    elif (dtype_x.kind == 'i') and (dtype_y.kind == 'f'):
        dtype_out = dtype_y.str
    else:
        dtype_out = 'O'

    return dtype_out



import numpy as np
from scipy.stats import t
# from scipy.stats.stats import _ttest_finish
from ..combinations import n_conditions_to_combinations
from ..array import samples_in_arr1_are_in_arr2, advanced_indexing
from . import descriptive as cc_stats_descriptive


def scores_raw_to_scores_formatted(scores_raw, axis_comparisons, dtype=None):

    shape_scores_raw = np.asarray(scores_raw.shape)
    n_axes_scores_raw = len(shape_scores_raw)

    if axis_comparisons < 0:
        axis_comparisons += n_axes_scores_raw

    axis_comparisons_raw = axis_comparisons

    scores_raw = np.expand_dims(scores_raw, axis=axis_comparisons_raw)
    scores_raw = np.expand_dims(scores_raw, axis=axis_comparisons_raw)
    axis_comparisons_raw += 2

    n_conditions_iv_per_comparisons = 2

    shape_scores_raw = np.asarray(scores_raw.shape)
    n_axes_scores_raw = len(shape_scores_raw)
    n_conditions_iv = shape_scores_raw[axis_comparisons_raw]
    conditions_iv = np.arange(n_conditions_iv)

    axes_scores = np.arange(n_axes_scores_raw)
    axes_non_independent_variables_scores_raw = axes_scores[np.logical_not(samples_in_arr1_are_in_arr2(
        axes_scores, axis_comparisons_raw))]

    axes_independent_variable_formatted = np.asarray([
        axis_comparisons_raw - 2,
        axis_comparisons_raw - 1,
        axis_comparisons_raw], dtype=int)

    axes_non_independent_variable_formatted = axes_scores[np.logical_not(samples_in_arr1_are_in_arr2(
        axes_scores, axes_independent_variable_formatted))]

    shape_scores_formatted = np.copy(shape_scores_raw)
    shape_scores_formatted[axes_independent_variable_formatted[0]] = n_conditions_iv_per_comparisons
    shape_scores_formatted[axes_independent_variable_formatted[1:]] = n_conditions_iv

    if dtype is None:
        dtype = scores_raw.dtype
    # scores_formatted = np.full(shape_scores_formatted, 0, dtype=dtype)
    scores_formatted = np.empty(shape_scores_formatted, dtype=dtype)

    index_0_raw = np.empty(n_axes_scores_raw, dtype=object)
    index_0_formatted = np.copy(index_0_raw)

    for a in axes_non_independent_variables_scores_raw:
        index_0_raw[a] = np.arange(shape_scores_raw[a])

    index_0_formatted[axes_non_independent_variable_formatted] = index_0_raw[
        axes_non_independent_variable_formatted]
    # index_0_formatted[axes_independent_variable_formatted[0]] = np.arange(n_conditions_iv_per_comparisons)

    index_1_formatted = np.copy(index_0_formatted)

    index_0_formatted[axes_independent_variable_formatted[0]] = 0
    index_0_formatted[axes_independent_variable_formatted[2]] = conditions_iv

    index_1_formatted[axes_independent_variable_formatted[0]] = 1
    index_1_formatted[axes_independent_variable_formatted[1]] = conditions_iv

    for i_condition in range(n_conditions_iv):

        index_0_raw[axis_comparisons_raw] = i_condition
        index_0_raw_adv = advanced_indexing(index_0_raw)

        scores_raw_0 = scores_raw[index_0_raw_adv]

        index_0_formatted[axes_independent_variable_formatted[1]] = i_condition
        index_0_formatted_adv = advanced_indexing(index_0_formatted)

        index_1_formatted[axes_independent_variable_formatted[2]] = i_condition
        index_1_formatted_adv = advanced_indexing(index_1_formatted)

        scores_formatted[index_0_formatted_adv] = scores_raw_0

        scores_formatted[index_1_formatted_adv] = scores_raw_0
        # print(scores_raw_0.shape)
        # print(scores_formatted[index_0_formatted_adv].shape)
        # print(scores_formatted[index_1_formatted_adv].shape)

    return scores_formatted


def scores_raw_to_paired_t_test(
        scores_raw, axis_samples, axis_comparisons, alpha=0.05, tails='2', keepdims=False):

    np.seterr(divide='ignore', invalid='ignore')

    shape_scores_raw = scores_raw.shape
    n_axes_scores_raw = len(shape_scores_raw)
    if axis_samples < 0:
        axis_samples += n_axes_scores_raw
    if axis_comparisons < 0:
        axis_comparisons += n_axes_scores_raw

    axis_comparisons_raw = axis_comparisons
    axis_samples_raw = axis_samples
    scores_formatted = scores_raw_to_scores_formatted(
        scores_raw, axis_comparisons=axis_comparisons_raw)

    if axis_samples_raw > axis_comparisons_raw:
        axis_samples_formatted = axis_samples_raw + 2
    else:
        axis_samples_formatted = axis_samples_raw

    axis_comparisons_formatted = axis_comparisons

    diff = cc_stats_descriptive.scores_to_diff_of_scores(
        scores_formatted, axis=axis_comparisons_formatted, keepdims=keepdims)

    if keepdims:
        axis_samples_diff = axis_samples_formatted
    else:
        axis_samples_diff = axis_samples_formatted - (axis_samples_formatted > axis_comparisons_formatted)

    n_samples = np.sum(np.logical_not(np.isnan(diff)), axis=axis_samples_diff, keepdims=keepdims)
    means_of_diff = cc_stats_descriptive.scores_to_means(
        diff, axes=axis_samples_diff, keepdims=keepdims)
    std_err_of_diff = cc_stats_descriptive.scores_to_standard_errors(
        diff, axis_samples=axis_samples_diff, keepdims=keepdims)

    t_values = means_of_diff / std_err_of_diff
    np.nan_to_num(t_values, copy=False, nan=0.0)

    confidence = 1 - alpha
    df = n_samples - 1
    shape_df = df.shape
    # n_df = df.size
    indexes_df = n_conditions_to_combinations(shape_df)
    t_critical = np.empty(shape_df, dtype=float)
    p_values = np.empty(shape_df, dtype=float)

    for indexes_df_i in indexes_df:
        indexes_df_i_tuple = tuple(indexes_df_i)
        if tails == '2':
            t_critical[indexes_df_i_tuple] = t.ppf((1 + confidence) / 2., df[indexes_df_i_tuple])
            p_values[indexes_df_i_tuple] = (1.0 - t.cdf(abs(t_values[indexes_df_i_tuple]), df[indexes_df_i_tuple])) * 2
        elif tails == '1l':
            t_critical[indexes_df_i_tuple] = -t.ppf(confidence, df[indexes_df_i_tuple])
            p_values[indexes_df_i_tuple] = t.cdf(t_values[indexes_df_i_tuple], df[indexes_df_i_tuple])
        elif tails == '1r':
            t_critical[indexes_df_i_tuple] = t.ppf(confidence, df[indexes_df_i_tuple])
            p_values[indexes_df_i_tuple] = 1.0 - t.cdf(t_values[indexes_df_i_tuple], df[indexes_df_i_tuple])

    # h = std_err_of_diff * t_critical

    return t_values, df, t_critical, p_values  # , h


def scores_formatted_to_paired_t_test(
        scores_formatted, axis_samples=-1, axis_comparisons=0, alpha=0.05, tails='2', keepdims=False):

    np.seterr(divide='ignore', invalid='ignore')

    shape_scores_formatted = scores_formatted.shape
    n_axes_scores_formatted = len(shape_scores_formatted)
    axis_samples_formatted = axis_samples % n_axes_scores_formatted
    axis_comparisons_formatted = axis_comparisons % n_axes_scores_formatted

    diff = cc_stats_descriptive.scores_to_diff_of_scores(
        scores_formatted, axis=axis_comparisons_formatted, keepdims=keepdims)

    if keepdims:
        axis_samples_diff = axis_samples_formatted
    else:
        axis_samples_diff = axis_samples_formatted - (axis_samples_formatted > axis_comparisons_formatted)

    n_samples = np.sum(np.logical_not(np.isnan(diff)), axis=axis_samples_diff, keepdims=keepdims)
    means_of_diff = cc_stats_descriptive.scores_to_means(
        diff, axes=axis_samples_diff, keepdims=keepdims)
    std_err_of_diff = cc_stats_descriptive.scores_to_standard_errors(
        diff, axis_samples=axis_samples_diff, keepdims=keepdims)

    t_values = means_of_diff / std_err_of_diff
    np.nan_to_num(t_values, copy=False, nan=0.0)

    confidence = 1 - alpha
    df = n_samples - 1
    shape_df = df.shape
    # n_df = df.size
    indexes_df = n_conditions_to_combinations(shape_df)
    t_critical = np.empty(shape_df, dtype=float)
    p_values = np.empty(shape_df, dtype=float)

    for indexes_df_i in indexes_df:
        indexes_df_i_tuple = tuple(indexes_df_i)
        if tails == '2':
            t_critical[indexes_df_i_tuple] = t.ppf((1 + confidence) / 2., df[indexes_df_i_tuple])
            p_values[indexes_df_i_tuple] = (1.0 - t.cdf(abs(t_values[indexes_df_i_tuple]), df[indexes_df_i_tuple])) * 2
        elif tails == '1l':
            t_critical[indexes_df_i_tuple] = -t.ppf(confidence, df[indexes_df_i_tuple])
            p_values[indexes_df_i_tuple] = t.cdf(t_values[indexes_df_i_tuple], df[indexes_df_i_tuple])
        elif tails == '1r':
            t_critical[indexes_df_i_tuple] = t.ppf(confidence, df[indexes_df_i_tuple])
            p_values[indexes_df_i_tuple] = 1.0 - t.cdf(t_values[indexes_df_i_tuple], df[indexes_df_i_tuple])

    # h = std_err_of_diff * t_critical

    return t_values, df, t_critical, p_values  # , h


def diff_to_paired_t_test(
        diff, axis_samples=-1, alpha=0.05, tails='2', keepdims=False):

    np.seterr(divide='ignore', invalid='ignore')

    shape_diff = diff.shape
    n_axes_diff = len(shape_diff)
    axis_samples_diff = axis_samples % n_axes_diff

    n_samples = np.sum(np.logical_not(np.isnan(diff)), axis=axis_samples_diff, keepdims=keepdims)
    means_of_diff = cc_stats_descriptive.scores_to_means(
        diff, axes=axis_samples_diff, keepdims=keepdims)
    std_err_of_diff = cc_stats_descriptive.scores_to_standard_errors(
        diff, axis_samples=axis_samples_diff, keepdims=keepdims)

    t_values = means_of_diff / std_err_of_diff
    np.nan_to_num(t_values, copy=False, nan=0.0)

    confidence = 1 - alpha
    df = n_samples - 1
    shape_df = df.shape
    # n_df = df.size
    indexes_df = n_conditions_to_combinations(shape_df)
    t_critical = np.empty(shape_df, dtype=float)
    p_values = np.empty(shape_df, dtype=float)

    for indexes_df_i in indexes_df:
        indexes_df_i_tuple = tuple(indexes_df_i)
        if tails == '2':
            t_critical[indexes_df_i_tuple] = t.ppf((1 + confidence) / 2., df[indexes_df_i_tuple])
            p_values[indexes_df_i_tuple] = (1.0 - t.cdf(abs(t_values[indexes_df_i_tuple]), df[indexes_df_i_tuple])) * 2
        elif tails == '1l':
            t_critical[indexes_df_i_tuple] = -t.ppf(confidence, df[indexes_df_i_tuple])
            p_values[indexes_df_i_tuple] = t.cdf(t_values[indexes_df_i_tuple], df[indexes_df_i_tuple])
        elif tails == '1r':
            t_critical[indexes_df_i_tuple] = t.ppf(confidence, df[indexes_df_i_tuple])
            p_values[indexes_df_i_tuple] = 1.0 - t.cdf(t_values[indexes_df_i_tuple], df[indexes_df_i_tuple])

    # h = std_err_of_diff * t_critical

    return t_values, df, t_critical, p_values  # , h



import numpy as np
from ...array import samples_in_arr1_are_in_arr2  # , advanced_indexing
from ...combinations import (
    trials_to_conditions, conditions_to_n_conditions,
    n_conditions_to_combinations, conditions_to_combinations)
from ..reshape import merge_axes_and_add_axes_removing_info_to_axis_receiving


def from_2_arrays(
        array_data_staying,
        array_indexes_adding_axes,
        axis_spitting_input,
        axis_indexes_adding_axes_input,
        axes_inserting_output=None,
        balance=True,
        dtype=None,
        format_and_check=True):

    shape_array_data_staying = np.asarray(array_data_staying.shape)
    n_axes_array_input = shape_array_data_staying.size

    if balance:

        n_indexes_staying = shape_array_data_staying[axis_indexes_adding_axes_input]

        if format_and_check:
            n_axes_array_data_staying = n_axes_array_input

            shape_array_indexes_adding_axes = np.asarray(array_indexes_adding_axes.shape)
            n_axes_array_indexes_adding_axes = shape_array_indexes_adding_axes.size

            # check point 1
            if n_axes_array_data_staying != n_axes_array_indexes_adding_axes:
                raise ValueError('dimension mismatch')

            if axis_indexes_adding_axes_input < 0:
                axis_indexes_adding_axes_input += n_axes_array_input
            if axis_spitting_input < 0:
                axis_spitting_input += n_axes_array_input

            # check point 5
            if axis_spitting_input == axis_indexes_adding_axes_input:
                raise ValueError('axis_spitting_input and axis_indexes_adding_axes_input must be different')

            # check point 2
            indexes_logical = np.arange(n_axes_array_input) != axis_indexes_adding_axes_input
            if np.any(shape_array_data_staying[indexes_logical] !=
                      shape_array_indexes_adding_axes[indexes_logical]):
                raise ValueError('dimension mismatch')

            n_indexes_table_adding_axes = shape_array_indexes_adding_axes[axis_indexes_adding_axes_input]

            # format axes_inserting_output
            n_axes_inserting_output = n_indexes_table_adding_axes
            n_axes_array_output = n_axes_array_input + n_axes_inserting_output
            try:
                len(axes_inserting_output)
                axes_inserting_output = np.asarray(axes_inserting_output, dtype=int)
                axes_inserting_output[axes_inserting_output < 0] += n_axes_array_output
                # check point 3
                for a in axes_inserting_output:
                    if np.sum(a == axes_inserting_output) > 1:
                        raise ValueError('axes_inserting_output cannot contain repeated values')
            except TypeError:
                if axes_inserting_output is None:
                    axes_inserting_output = np.arange(n_axes_inserting_output)
                else:
                    if axes_inserting_output < 0:
                        axes_inserting_output += n_axes_array_output
                    axes_inserting_output = np.asarray([axes_inserting_output], dtype=int)

        else:
            n_axes_inserting_output = axes_inserting_output.size
            n_axes_array_output = n_axes_array_input + n_axes_inserting_output

        axes_array_input = np.arange(n_axes_array_input)
        axes_non_axis_indexes_adding_axes_input = axes_array_input[axes_array_input != axis_indexes_adding_axes_input]
        axes_other_array_input = axes_non_axis_indexes_adding_axes_input[
            axes_non_axis_indexes_adding_axes_input != axis_spitting_input]
        n_axes_other_array_input = axes_other_array_input.size

        indexes_array_data_staying = np.empty(n_axes_array_input, dtype=object)
        indexes_array_data_staying[:] = slice(None)

        axis_variables_in_combinations = int(axis_indexes_adding_axes_input > axis_spitting_input)
        axis_combinations_in_combinations = int(not (bool(axis_variables_in_combinations)))

        if n_axes_other_array_input == 0:
            conditions_indexes_table_adding_axes = trials_to_conditions(
                array_indexes_adding_axes, axis_combinations=axis_combinations_in_combinations)
        elif n_axes_other_array_input > 0:
            indexes_array_indexes_adding_axes = np.copy(indexes_array_data_staying)
            indexes_array_indexes_adding_axes[axes_other_array_input] = 0
            array_indexes_adding_axes_1 = array_indexes_adding_axes[tuple(indexes_array_indexes_adding_axes)]
            conditions_indexes_table_adding_axes = trials_to_conditions(
                array_indexes_adding_axes_1, axis_combinations=axis_combinations_in_combinations)

        n_conditions_indexes_table_adding_axes = conditions_to_n_conditions(
            conditions_indexes_table_adding_axes)
        n_combinations_indexes_table_adding_axes = int(np.prod(
            n_conditions_indexes_table_adding_axes))

        combinations_indexes_table_adding_axes = conditions_to_combinations(
            conditions_indexes_table_adding_axes,
            axis_combinations=axis_combinations_in_combinations)

        combinations_axes_inserting_output = n_conditions_to_combinations(
            n_conditions_indexes_table_adding_axes)

        axis_indexes_removing_output = axis_indexes_adding_axes_input + np.sum(
            axes_inserting_output <= axis_indexes_adding_axes_input)
        axis_spitting_output = axis_spitting_input + np.sum(axes_inserting_output <= axis_spitting_input)
        need_to_check = True
        while need_to_check:
            need_to_check = False
            while axis_indexes_removing_output in axes_inserting_output:
                axis_indexes_removing_output += 1
                need_to_check = True
            while axis_spitting_output in axes_inserting_output:
                axis_spitting_output += 1
                need_to_check = True
            if axis_indexes_removing_output == axis_spitting_output:
                if axis_indexes_adding_axes_input > axis_spitting_input:
                    axis_indexes_removing_output += 1
                elif axis_indexes_adding_axes_input < axis_spitting_input:
                    axis_spitting_output += 1
                need_to_check = True

        axes_array_output = np.arange(n_axes_array_output)
        axes_non_axis_indexes_removing_output = axes_array_output[axes_array_output != axis_indexes_removing_output]
        axes_other_output = axes_non_axis_indexes_removing_output[
            axes_non_axis_indexes_removing_output != axis_spitting_output]
        axes_other_output = axes_other_output[np.logical_not(
            samples_in_arr1_are_in_arr2(axes_other_output, axes_inserting_output))]
        # axes_inserting_output_sorted = np.sort(axes_inserting_output)

        shape_array_output = np.empty(n_axes_array_output, dtype=int)
        shape_array_output[axes_inserting_output] = n_conditions_indexes_table_adding_axes
        shape_array_output[axis_indexes_removing_output] = n_indexes_staying
        shape_array_output[axis_spitting_output] = (
                shape_array_data_staying[axis_spitting_input] / n_combinations_indexes_table_adding_axes)
        shape_array_output[axes_other_output] = shape_array_data_staying[axes_other_array_input]
        if dtype is None:
            dtype = array_data_staying.dtype
        array_output = np.empty(shape_array_output, dtype=dtype)

        indexes_output = np.empty(n_axes_array_output, dtype=object)
        indexes_output[:] = slice(None)

        indexes_combinations = np.empty(2, dtype=object)
        indexes_combinations[axis_variables_in_combinations] = slice(None)

        if n_axes_other_array_input == 0:
            for i in range(n_combinations_indexes_table_adding_axes):
                indexes_output[axes_inserting_output] = combinations_axes_inserting_output[i]
                indexes_combinations[axis_combinations_in_combinations] = slice(i, i + 1)
                indexes_array_data_staying[axis_spitting_input] = np.all(
                    array_indexes_adding_axes ==
                    combinations_indexes_table_adding_axes[tuple(indexes_combinations)],
                    axis=axis_variables_in_combinations)
                array_output[tuple(indexes_output)] = array_data_staying[tuple(
                    indexes_array_data_staying)]

        elif n_axes_other_array_input > 0:
            indexes_array_data_staying_j = np.empty(2, dtype=object)
            indexes_array_data_staying_j[axis_variables_in_combinations] = slice(None)

            combinations_axes_other = n_conditions_to_combinations(
                shape_array_data_staying[axes_other_array_input])
            n_combinations_axes_other = combinations_axes_other.shape[0]
            for j in range(n_combinations_axes_other):
                combinations_axes_other_j = combinations_axes_other[j]
                indexes_output[axes_other_output] = combinations_axes_other_j
                indexes_array_data_staying[axes_other_array_input] = combinations_axes_other_j
                indexes_array_indexes_adding_axes[axes_other_array_input] = combinations_axes_other_j
                array_indexes_adding_axes_j = array_indexes_adding_axes[
                    tuple(indexes_array_indexes_adding_axes)]
                array_data_staying_j = array_data_staying[tuple(
                    indexes_array_data_staying)]
                for i in range(n_combinations_indexes_table_adding_axes):
                    indexes_output[axes_inserting_output] = combinations_axes_inserting_output[i]
                    indexes_combinations[axis_combinations_in_combinations] = slice(i, i + 1)
                    indexes_array_data_staying_j[axis_combinations_in_combinations] = np.all(
                        array_indexes_adding_axes_j ==
                        combinations_indexes_table_adding_axes[tuple(indexes_combinations)],
                        axis=axis_variables_in_combinations)
                    array_output[tuple(indexes_output)] = array_data_staying_j[tuple(
                        indexes_array_data_staying_j)]

    else:
        # n_indexes_staying = shape_array_data_staying[axis_indexes_adding_axes_input]
        axes_array_input = np.arange(n_axes_array_input)
        axes_non_axis_indexes_adding_axes_input = axes_array_input[axes_array_input != axis_indexes_adding_axes_input]
        axes_other_array_input = axes_non_axis_indexes_adding_axes_input[
            axes_non_axis_indexes_adding_axes_input != axis_spitting_input]
        n_axes_other_array_input = axes_other_array_input.size

        if format_and_check:
            n_axes_array_data_staying = n_axes_array_input

            shape_array_indexes_adding_axes = np.asarray(array_indexes_adding_axes.shape)
            n_axes_array_indexes_adding_axes = shape_array_indexes_adding_axes.size

            # check point 1
            if n_axes_array_data_staying != n_axes_array_indexes_adding_axes:
                raise ValueError('dimension mismatch')

            if axis_indexes_adding_axes_input < 0:
                axis_indexes_adding_axes_input += n_axes_array_input
            if axis_spitting_input < 0:
                axis_spitting_input += n_axes_array_input

            # check point 5
            if axis_spitting_input == axis_indexes_adding_axes_input:
                raise ValueError('axis_spitting_input and axis_indexes_adding_axes_input must be different')

            # check point 2
            indexes_logical = np.arange(n_axes_array_input) != axis_indexes_adding_axes_input
            if np.any(shape_array_data_staying[indexes_logical] !=
                      shape_array_indexes_adding_axes[indexes_logical]):
                raise ValueError('dimension mismatch')

            n_indexes_table_adding_axes = shape_array_indexes_adding_axes[axis_indexes_adding_axes_input]

            # format axes_inserting_output
            n_axes_inserting_output = n_indexes_table_adding_axes
            n_axes_array_output_object = n_axes_inserting_output
            try:
                len(axes_inserting_output)
                axes_inserting_output = np.asarray(axes_inserting_output, dtype=int)
                axes_inserting_output[axes_inserting_output < 0] += n_axes_array_output_object
                # check point 3
                for a in axes_inserting_output:
                    if np.sum(a == axes_inserting_output) > 1:
                        raise ValueError('axes_inserting_output cannot contain repeated values')

            except TypeError:
                if axes_inserting_output is None:
                    axes_inserting_output = np.arange(n_axes_inserting_output)
                else:
                    if axes_inserting_output < 0:
                        axes_inserting_output += n_axes_array_output_object
                    axes_inserting_output = np.asarray([axes_inserting_output], dtype=int)

            if n_axes_other_array_input == 0:
                if dtype is not None and dtype != array_data_staying.dtype:
                    array_data_staying = array_data_staying.astype(dtype)
            else:
                if dtype is None:
                    dtype = array_data_staying.dtype
        else:
            n_axes_inserting_output = axes_inserting_output.size
            n_axes_array_output_object = n_axes_inserting_output

        indexes_array_data_staying = np.empty(n_axes_array_input, dtype=object)
        indexes_array_data_staying[:] = slice(None)

        axis_variables_in_combinations = int(axis_indexes_adding_axes_input > axis_spitting_input)
        axis_combinations_in_combinations = int(not (bool(axis_variables_in_combinations)))

        if n_axes_other_array_input == 0:
            conditions_indexes_table_adding_axes = trials_to_conditions(
                array_indexes_adding_axes, axis_combinations=axis_combinations_in_combinations)
        elif n_axes_other_array_input > 0:
            # ---------------------------------------------------------------------------------------------
            # FAST BUT MAYBE WRONG IN SOME CASES 
            indexes_array_indexes_adding_axes = np.copy(indexes_array_data_staying)
            # indexes_array_indexes_adding_axes[axes_other_array_input] = 0
            # array_indexes_adding_axes_1 = array_indexes_adding_axes[tuple(indexes_array_indexes_adding_axes)]
            # conditions_indexes_table_adding_axes = trials_to_conditions(
            #     array_indexes_adding_axes_1, axis_combinations=axis_combinations_in_combinations)
            # ----------------------------------------------------------------------------------------
            
            # ---------------------------------------------------------------------------------------
            # SLOW BUT SAFIER IN SOME CASES
            array_indexes_adding_axes_flat = merge_axes_and_add_axes_removing_info_to_axis_receiving(
                array_indexes_adding_axes,
                axes_other_array_input,
                axis_spitting_input,
                axis_indexes_adding_axes_input,
                indexes_inserting_output=np.arange(n_axes_other_array_input),
                dtype=array_indexes_adding_axes.dtype)

            indexes = np.empty(2, dtype=object)
            indexes[axis_combinations_in_combinations] = slice(0, None, 1)
            indexes[axis_variables_in_combinations] = slice(n_axes_other_array_input, None, 1)
            array_indexes_removing_object = array_indexes_adding_axes_flat[tuple(indexes)]
            # indexes[axis_variables_in_combinations] = slice(0, n_axes_other_array_input, 1)
            # array_indexes_removing_not_object = array_indexes_adding_axes_flat[tuple(indexes)]
            conditions_indexes_table_adding_axes = trials_to_conditions(
                array_indexes_removing_object, axis_combinations=axis_combinations_in_combinations)

            # conditions_indexes_removing_not_object = trials_to_conditions(
            #     array_indexes_removing_not_object, axis_combinations=axis_combinations_in_combinations)
            # ----------------------------------------------------------------------------------------

        n_conditions_indexes_table_adding_axes = conditions_to_n_conditions(
            conditions_indexes_table_adding_axes)
        n_combinations_indexes_table_adding_axes = int(np.prod(
            n_conditions_indexes_table_adding_axes))

        combinations_indexes_table_adding_axes = conditions_to_combinations(
            conditions_indexes_table_adding_axes,
            axis_combinations=axis_combinations_in_combinations)

        combinations_axes_inserting_output = n_conditions_to_combinations(
            n_conditions_indexes_table_adding_axes)

        # axes_array_output_object = np.arange(n_axes_array_output_object)
        shape_array_output_object = np.empty(n_axes_array_output_object, dtype=int)
        shape_array_output_object[axes_inserting_output] = n_conditions_indexes_table_adding_axes
        array_output = np.empty(shape_array_output_object, dtype=object)

        indexes_output_object = np.empty(n_axes_array_output_object, dtype=object)

        indexes_combinations = np.empty(2, dtype=object)
        indexes_combinations[axis_variables_in_combinations] = slice(None)

        if n_axes_other_array_input == 0:

            for i in range(n_combinations_indexes_table_adding_axes):
                indexes_output_object[axes_inserting_output] = combinations_axes_inserting_output[i]
                indexes_combinations[axis_combinations_in_combinations] = slice(i, i + 1)
                indexes_array_data_staying[axis_spitting_input] = np.all(
                    array_indexes_adding_axes ==
                    combinations_indexes_table_adding_axes[tuple(indexes_combinations)],
                    axis=axis_variables_in_combinations)
                array_output[tuple(indexes_output_object)] = array_data_staying[tuple(
                    indexes_array_data_staying)]

        elif n_axes_other_array_input > 0:

            shape_array_output = np.copy(shape_array_data_staying)

            indexes_output = np.empty(n_axes_array_input, dtype=object)
            indexes_output[:] = slice(None)
            indexes_array_data_staying_j = np.empty(2, dtype=object)
            indexes_array_data_staying_j[axis_variables_in_combinations] = slice(None)

            combinations_axes_other_0 = np.empty(n_axes_other_array_input, dtype=int)
            combinations_axes_other_0[:] = 0
            indexes_output[axes_other_array_input] = combinations_axes_other_0
            indexes_output_tuple = tuple(indexes_output)
            indexes_array_data_staying[axes_other_array_input] = combinations_axes_other_0
            indexes_array_indexes_adding_axes[axes_other_array_input] = combinations_axes_other_0
            array_indexes_adding_axes_0 = array_indexes_adding_axes[
                tuple(indexes_array_indexes_adding_axes)]
            array_data_staying_0 = array_data_staying[tuple(
                indexes_array_data_staying)]
            for i in range(n_combinations_indexes_table_adding_axes):
                indexes_output_object[axes_inserting_output] = combinations_axes_inserting_output[i]
                indexes_combinations[axis_combinations_in_combinations] = slice(i, i + 1)
                indexes_array_data_staying_j[axis_combinations_in_combinations] = np.all(
                    array_indexes_adding_axes_0 ==
                    combinations_indexes_table_adding_axes[tuple(indexes_combinations)],
                    axis=axis_variables_in_combinations)

                shape_array_output[axis_spitting_input] = np.sum(
                    indexes_array_data_staying_j[axis_combinations_in_combinations])

                array_output[tuple(indexes_output_object)] = np.empty(shape_array_output, dtype=dtype)

                array_output[tuple(indexes_output_object)][indexes_output_tuple] = array_data_staying_0[
                    tuple(indexes_array_data_staying_j)]

            combinations_axes_other = n_conditions_to_combinations(
                shape_array_data_staying[axes_other_array_input])
            n_combinations_axes_other = combinations_axes_other.shape[0]

            for j in range(1, n_combinations_axes_other):
                combinations_axes_other_j = combinations_axes_other[j]
                indexes_output[axes_other_array_input] = combinations_axes_other_j
                indexes_output_tuple = tuple(indexes_output)
                indexes_array_data_staying[axes_other_array_input] = combinations_axes_other_j
                indexes_array_indexes_adding_axes[axes_other_array_input] = combinations_axes_other_j
                array_indexes_adding_axes_j = array_indexes_adding_axes[
                    tuple(indexes_array_indexes_adding_axes)]
                array_data_staying_j = array_data_staying[tuple(
                    indexes_array_data_staying)]
                for i in range(n_combinations_indexes_table_adding_axes):
                    indexes_output_object[axes_inserting_output] = combinations_axes_inserting_output[i]
                    indexes_combinations[axis_combinations_in_combinations] = slice(i, i + 1)
                    indexes_array_data_staying_j[axis_combinations_in_combinations] = np.all(
                        array_indexes_adding_axes_j ==
                        combinations_indexes_table_adding_axes[tuple(indexes_combinations)],
                        axis=axis_variables_in_combinations)
                    array_output[tuple(indexes_output_object)][indexes_output_tuple] = array_data_staying_j[
                        tuple(indexes_array_data_staying_j)]

    return array_output


def from_2_arrays_advanced(
        array_data_staying,
        array_indexes_adding_axes,
        axis_spitting_input,
        axis_indexes_adding_axes_input,
        indexes_table_adding_axes=None,
        axes_inserting_output=None,
        indexes_staying=None,
        balance=True,
        dtype=None,
        format_and_check=True):

    # Notes:
    # 1) uses the efficient function slice for indexes;
    # 2) it splits the array in two arrays first: array_indexes_adding_axes and array_data_staying.
    #    Then, it makes the array output from those two arrays;
    # 3) it assumes that the numbers of samples (or trials) in the axis_spitting_input for all combinations
    #    of indexes_table_adding_axes are equal;
    # 4) it does not assumes that the order of samples (or trials) in the axis_spitting_input are is the same
    #    in each combination of variables' axes. The order is based on the order of the combinations
    #    of variables' conditions in axis_indexes_adding_axes_input. In other words, the samples can be in random order
    #    (or in different order) in each combinations of variables axes/conditions.

    # Input requirements:
    # 1) the numbers of axes of the two arrays must be equal;
    # 2) the shapes of the 2 arrays must be equal, except for the axis_variable_table,
    #    i.e. the axis that contains the variables of the tables;
    # 3) axes_inserting_output cannot contain repeated values;
    # 4) indexes_table_adding_axes cannot contain repeated values;
    # 5) shapes of axes_inserting_output and indexes_table_adding_axes must be equal
    # 6) axis_spitting_input != axis_indexes_adding_axes_input
    # 7) the numbers of samples (or trials) in the axis_spitting_input for all combinations
    #    of indexes_table_adding_axes are equal

    # Import requirements:
    # 1) import numpy as np
    # 2) from ccalafiore.combinations import n_conditions_to_combinations
    # 3) from ccalafiore.combinations import trials_to_conditions
    # 4) from ccalafiore.combinations import conditions_to_n_conditions
    # 5) from ccalafiore.combinations import conditions_to_combinations
    # 6) from ccalafiore.array import samples_in_arr1_are_in_arr2

    # from ccalafiore.combinations import \
    #     trials_to_conditions, conditions_to_n_conditions, n_conditions_to_combinations, conditions_to_combinations
    # from ccalafiore.array import samples_in_arr1_are_in_arr2

    shape_array_data_staying = np.asarray(array_data_staying.shape)
    n_axes_array_data_staying = shape_array_data_staying.size
    n_variables_table_in_array_data_staying = shape_array_data_staying[axis_indexes_adding_axes_input]
    variables_table_in_array_data_staying = np.arange(n_variables_table_in_array_data_staying)

    shape_array_indexes_adding_axes = np.asarray(array_indexes_adding_axes.shape)
    n_axes_array_indexes_adding_axes = shape_array_indexes_adding_axes.size
    n_variables_table_in_array_indexes_adding_axes = shape_array_indexes_adding_axes[axis_indexes_adding_axes_input]
    variables_table_in_array_indexes_adding_axes = np.arange(n_variables_table_in_array_indexes_adding_axes)

    if format_and_check:
        # check point 1
        if n_axes_array_data_staying != n_axes_array_indexes_adding_axes:
            raise ValueError('dimension mismatch')
        else:
            n_axes_array_input = n_axes_array_data_staying

        if axis_indexes_adding_axes_input < 0:
            axis_indexes_adding_axes_input += n_axes_array_input
        if axis_spitting_input < 0:
            axis_spitting_input += n_axes_array_input

        # check point 6
        if axis_spitting_input == axis_indexes_adding_axes_input:
            raise ValueError('axis_spitting_input and axis_indexes_adding_axes_input must be different')

        # check point 2
        indexes_logical = np.arange(n_axes_array_input) != axis_indexes_adding_axes_input
        if np.any(shape_array_data_staying[indexes_logical] != shape_array_indexes_adding_axes[indexes_logical]):
            raise ValueError('dimension mismatch')

        n_variables_table_in_array_data_staying = shape_array_data_staying[axis_indexes_adding_axes_input]
        variables_table_in_array_data_staying = np.arange(n_variables_table_in_array_data_staying)

        n_variables_table_in_array_indexes_adding_axes = shape_array_indexes_adding_axes[axis_indexes_adding_axes_input]
        variables_table_in_array_indexes_adding_axes = np.arange(n_variables_table_in_array_indexes_adding_axes)

        # format indexes_table_adding_axes
        try:
            n_indexes_table_adding_axes = len(indexes_table_adding_axes)
            indexes_table_adding_axes = np.asarray(indexes_table_adding_axes, dtype=int)
            indexes_table_adding_axes[indexes_table_adding_axes < 0] += \
                n_variables_table_in_array_indexes_adding_axes
            # check point 4
            for v in indexes_table_adding_axes:
                if np.sum(v == indexes_table_adding_axes) > 1:
                    raise ValueError('indexes_table_adding_axes cannot contain repeated values')
        except TypeError:
            if indexes_table_adding_axes is None:
                indexes_table_adding_axes = np.arange(n_variables_table_in_array_indexes_adding_axes)
                n_indexes_table_adding_axes = n_variables_table_in_array_indexes_adding_axes
            else:
                if indexes_table_adding_axes < 0:
                    indexes_table_adding_axes += n_variables_table_in_array_indexes_adding_axes
                indexes_table_adding_axes = np.asarray([indexes_table_adding_axes], dtype=int)
                n_indexes_table_adding_axes = 1

        # format indexes_staying
        try:
            n_indexes_staying = len(indexes_staying)
            indexes_staying = np.asarray(indexes_staying, dtype=int)
            indexes_staying[indexes_staying < 0] += n_variables_table_in_array_data_staying
        except TypeError:
            if indexes_staying is None:
                indexes_staying = np.arange(n_variables_table_in_array_data_staying)
                n_indexes_staying = n_variables_table_in_array_data_staying
            else:
                if indexes_staying < 0:
                    indexes_staying += n_variables_table_in_array_data_staying
                indexes_staying = np.asarray([indexes_staying], dtype=int)
                n_indexes_staying = 1

        # format axes_inserting_output
        n_axes_inserting_output = n_indexes_table_adding_axes
        if balance:
            n_axes_array_output = n_axes_array_input + n_axes_inserting_output
        else:
            n_axes_array_output = n_axes_inserting_output
        try:
            n_axes_inserting_output = len(axes_inserting_output)
            axes_inserting_output = np.asarray(axes_inserting_output, dtype=int)
            axes_inserting_output[axes_inserting_output < 0] += n_axes_array_output
            # check point 3
            for a in axes_inserting_output:
                if np.sum(a == axes_inserting_output) > 1:
                    raise ValueError('axes_inserting_output cannot contain repeated values')
            # check point 5
            if n_indexes_table_adding_axes != n_axes_inserting_output:
                raise ValueError('Shapes of axes_inserting_output and indexes_table_adding_axes must be equal')
        except TypeError:
            if axes_inserting_output is None:
                axes_inserting_output = np.arange(n_axes_inserting_output)
            else:
                if axes_inserting_output < 0:
                    axes_inserting_output += n_axes_array_output
                axes_inserting_output = np.asarray([axes_inserting_output], dtype=int)
                n_axes_inserting_output = 1
                # check point 5
                if n_indexes_table_adding_axes != n_axes_inserting_output:
                    raise ValueError('Shapes of axes_inserting_output and indexes_table_adding_axes must be equal')
    else:
        n_indexes_staying = indexes_staying.size
        n_indexes_table_adding_axes = indexes_table_adding_axes.size

    indexes = None
    if ((n_indexes_staying != n_variables_table_in_array_data_staying) or
            np.any(indexes_staying != variables_table_in_array_data_staying)):

        indexes = np.empty(n_axes_array_data_staying, dtype=object)
        indexes[:] = slice(None)
        indexes[axis_indexes_adding_axes_input] = indexes_staying
        array_data_staying = array_data_staying[tuple(indexes)]

    if ((n_indexes_table_adding_axes != n_variables_table_in_array_indexes_adding_axes) or
            np.any(indexes_table_adding_axes != variables_table_in_array_indexes_adding_axes)):

        if indexes is None:
            indexes = np.empty(n_axes_array_indexes_adding_axes, dtype=object)
            indexes[:] = slice(None)
        indexes[axis_indexes_adding_axes_input] = indexes_table_adding_axes
        array_indexes_adding_axes = array_indexes_adding_axes[tuple(indexes)]

    array_output = from_2_arrays(
        array_data_staying,
        array_indexes_adding_axes,
        axis_spitting_input,
        axis_indexes_adding_axes_input,
        axes_inserting_output=axes_inserting_output,
        balance=balance,
        dtype=dtype,
        format_and_check=False)

    return array_output


def from_1_array(
        array_input,
        axis_spitting_input,
        axis_indexes_adding_axes_input,
        indexes_table_adding_axes,
        axes_inserting_output=None,
        indexes_staying=None,
        balance=True,
        dtype=None,
        format_and_check=True):

    # Notes:
    # 1) uses the efficient function slice for indexes;
    # 2) it splits the array in two arrays first: array_indexes_adding_axes and array_data_staying.
    #    Then, it makes the array output from those two arrays;
    # 3) the numbers of samples (or trials) in the axis_spitting_input for all combinations
    #    of indexes_table_adding_axes have to be equal;
    # 4) it does not assumes that the order of samples (or trials) in the axis_spitting_input are is the same
    #    in each combination of variables' axes. The order is based on the order of the combinations
    #    of variables' conditions in axis_indexes_adding_axes_input. In other words, the samples can be in random order
    #    (or in different order) in each combinations of variables axes/conditions.

    # Input requirements:
    # 1) axes_inserting_output cannot contain repeated values;
    # 2) indexes_table_adding_axes cannot contain repeated values;
    # 3) shapes of axes_inserting_output and indexes_table_adding_axes must be equal
    # 4) axis_spitting_input != axis_indexes_adding_axes_input
    # 5) the numbers of samples (or trials) in the axis_spitting_input for all combinations
    #    of indexes_table_adding_axes are equal

    # Import requirements:
    # 1) import numpy as np
    # 2) from ccalafiore.combinations import n_conditions_to_combinations
    # 3) from ccalafiore.combinations import trials_to_conditions
    # 4) from ccalafiore.combinations import conditions_to_n_conditions
    # 5) from ccalafiore.combinations import conditions_to_combinations
    # 6) from ccalafiore.array import samples_in_arr1_are_in_arr2

    # from ccalafiore.combinations import \
    #     trials_to_conditions, conditions_to_n_conditions, n_conditions_to_combinations, conditions_to_combinations
    # from ccalafiore.array import samples_in_arr1_are_in_arr2

    shape_array_input = np.asarray(array_input.shape)
    n_axes_array_input = shape_array_input.size
    n_variables_table_input = shape_array_input[axis_indexes_adding_axes_input]
    variables_table_input = np.arange(n_variables_table_input)

    if format_and_check:
        if axis_indexes_adding_axes_input < 0:
            axis_indexes_adding_axes_input += n_axes_array_input
        if axis_spitting_input < 0:
            axis_spitting_input += n_axes_array_input

        # check point 4
        if axis_spitting_input == axis_indexes_adding_axes_input:
            raise ValueError('axis_spitting_input and axis_indexes_adding_axes_input must be different')

        # format indexes_table_adding_axes
        try:
            n_indexes_table_adding_axes = len(indexes_table_adding_axes)
            indexes_table_adding_axes = np.asarray(indexes_table_adding_axes, dtype=int)
            indexes_table_adding_axes[indexes_table_adding_axes < 0] += n_variables_table_input
            # check point 2
            for v in indexes_table_adding_axes:
                if np.sum(v == indexes_table_adding_axes) > 1:
                    raise ValueError('indexes_table_adding_axes cannot contain repeated values')
        except TypeError:
            if indexes_table_adding_axes < 0:
                indexes_table_adding_axes += n_variables_table_input
            indexes_table_adding_axes = np.asarray([indexes_table_adding_axes], dtype=int)
            n_indexes_table_adding_axes = 1

        # format indexes_staying
        try:
            n_indexes_staying = len(indexes_staying)
            indexes_staying = np.asarray(indexes_staying, dtype=int)
            indexes_staying[indexes_staying < 0] += n_variables_table_input
        except TypeError:
            if indexes_staying is None:
                indexes_staying = variables_table_input[np.logical_not(
                    samples_in_arr1_are_in_arr2(variables_table_input, indexes_table_adding_axes))]
                n_indexes_staying = indexes_staying.size
            else:
                if indexes_staying < 0:
                    indexes_staying += n_variables_table_input
                indexes_staying = np.asarray([indexes_staying], dtype=int)
                n_indexes_staying = 1

        # format axes_inserting_output
        n_axes_inserting_output = n_indexes_table_adding_axes
        if balance:
            n_axes_array_output = n_axes_array_input + n_axes_inserting_output
        else:
            n_axes_array_output = n_axes_inserting_output
        try:
            n_axes_inserting_output = len(axes_inserting_output)
            axes_inserting_output = np.asarray(axes_inserting_output, dtype=int)
            axes_inserting_output[axes_inserting_output < 0] += n_axes_array_output
            # check point 1
            for a in axes_inserting_output:
                if np.sum(a == axes_inserting_output) > 1:
                    raise ValueError('axes_inserting_output cannot contain repeated values')
            # check point 3
            if n_indexes_table_adding_axes != n_axes_inserting_output:
                raise ValueError('Shapes of axes_inserting_output and indexes_table_adding_axes must be equal')
        except TypeError:
            if axes_inserting_output is None:
                axes_inserting_output = np.arange(n_axes_inserting_output)
            else:
                if axes_inserting_output < 0:
                    axes_inserting_output += n_axes_array_output
                axes_inserting_output = np.asarray([axes_inserting_output], dtype=int)
                n_axes_inserting_output = 1
                # check point 3
                if n_indexes_table_adding_axes != n_axes_inserting_output:
                    raise ValueError('Shapes of axes_inserting_output and indexes_table_adding_axes must be equal')
    else:
        n_indexes_staying = indexes_staying.size
        n_indexes_table_adding_axes = indexes_table_adding_axes.size

    indexes = np.empty(n_axes_array_input, dtype=object)
    indexes[:] = slice(None)
    if ((n_indexes_staying != n_variables_table_input) or
            np.any(indexes_staying != variables_table_input)):
        indexes[axis_indexes_adding_axes_input] = indexes_staying
    array_data_staying = array_input[tuple(indexes)]

    if ((n_indexes_table_adding_axes != n_variables_table_input) or
            np.any(indexes_table_adding_axes != variables_table_input)):
        indexes[axis_indexes_adding_axes_input] = indexes_table_adding_axes
    else:
        indexes[axis_indexes_adding_axes_input] = slice(None)
    array_indexes_adding_axes = array_input[tuple(indexes)]
    del array_input

    array_output = from_2_arrays(
        array_data_staying,
        array_indexes_adding_axes,
        axis_spitting_input,
        axis_indexes_adding_axes_input,
        axes_inserting_output=axes_inserting_output,
        balance=balance,
        dtype=dtype,
        format_and_check=False)

    return array_output



import numpy as np
from ...array import samples_in_arr1_are_in_arr2
from .balanced_or_unbalanced import from_2_arrays as v2a_from_2_arrays
# from .balanced import from_2_arrays as v2a_balanced_from_2_arrays
# from .unbalanced import from_2_arrays as v2a_unbalanced_from_2_arrays


def from_3_arrays(
        array_variables_staying,
        array_variables_adding_axes_balanced,
        array_variables_adding_axes_unbalanced,
        axis_samples,
        axis_variables_table,
        axes_inserting_array_not_object=None,
        axes_inserting_array_object=None,
        dtype=None,
        format_and_check=True):

    # Notes:
    # 1) it is safe;
    # 2) it does not assumes that the numbers of samples (or trials) in the axis_samples for all
    #    combinations of variables_table_adding_axes_balanced are equal;
    # 3) it does not assumes that the order of samples (or trials) in the axis_samples are is the same
    #    in each combination of variables' axes. The order is based on the order of the combinations
    #    of variables' conditions in axis_variables_table. In other words, the samples can be in random order
    #    (or in different order) in each combinations of variables axes/conditions.

    # Input requirements:
    # 1) axes_inserting_array_not_object cannot contain repeated values;
    # 2) axes_inserting_array_object cannot contain repeated values;
    # 3) axis_samples != axis_variables_table
    # 4) the numbers of axes of the 3 arrays must be equal;
    # 5) the shapes of the 3 arrays must be equal, except for the axis_variable_table,
    #     i.e. the axis that contains the variables of the tables;
    # 6) Either one of the 3 conditions must be True:
    #    array_variables_adding_axes_balanced != None;
    #    array_variables_adding_axes_unbalanced != None;
    #    array_variables_adding_axes_balanced != None and array_variables_adding_axes_unbalanced != None;

    # Import requirements:
    # 1) import numpy as np
    # 2) from ccalafiore.combinations import n_conditions_to_combinations
    # 3) from ccalafiore.combinations import trials_to_conditions
    # 4) from ccalafiore.combinations import conditions_to_n_conditions
    # 5) from ccalafiore.combinations import conditions_to_combinations
    # 6) from ccalafiore.array import samples_in_arr1_are_in_arr2

    # from ccalafiore.combinations import \
    #     trials_to_conditions, conditions_to_n_conditions, n_conditions_to_combinations, conditions_to_combinations
    # from ccalafiore.array import samples_in_arr1_are_in_arr2

    # check point 6
    if array_variables_adding_axes_balanced is None and array_variables_adding_axes_unbalanced is None:
        raise ValueError('Either one of the 3 conditions must be True:'
                         'array_variables_adding_axes_balanced != None\n'
                         'array_variables_adding_axes_unbalanced != None\n'
                         'array_variables_adding_axes_balanced != None and '
                         'array_variables_adding_axes_unbalanced != None;')

    axis_variables_table_input = axis_variables_table
    axis_samples_input = axis_samples

    if format_and_check:

        shape_array_variables_staying = np.asarray(array_variables_staying.shape)
        shape_array_variables_adding_axes_balanced = np.asarray(array_variables_adding_axes_balanced.shape)
        shape_array_variables_adding_axes_unbalanced = np.asarray(array_variables_adding_axes_unbalanced.shape)

        # check point 4
        n_axes_array_variables_staying = shape_array_variables_staying.size
        n_axes_array_variables_adding_axes_balanced = shape_array_variables_adding_axes_balanced.size
        n_axes_array_variables_adding_axes_unbalanced = shape_array_variables_adding_axes_unbalanced.size
        if (n_axes_array_variables_staying != n_axes_array_variables_adding_axes_balanced) or (
                n_axes_array_variables_staying != n_axes_array_variables_adding_axes_unbalanced):
            raise ValueError('dimension mismatch')
        else:
            n_axes_array_input = n_axes_array_variables_staying

        if axis_variables_table_input < 0:
            axis_variables_table_input += n_axes_array_input
        if axis_samples_input < 0:
            axis_samples_input += n_axes_array_input

        # check point 3
        if axis_samples_input == axis_variables_table_input:
            raise ValueError('axis_samples_input and axis_variables_table_input must be different')

        # check point 5
        indexes_logical = np.arange(n_axes_array_input) != axis_variables_table_input
        if np.any(shape_array_variables_staying[indexes_logical] !=
                  shape_array_variables_adding_axes_balanced[indexes_logical]):
            raise ValueError('dimension mismatch')
        if np.any(shape_array_variables_staying[indexes_logical] !=
                  shape_array_variables_adding_axes_unbalanced[indexes_logical]):
            raise ValueError('dimension mismatch')

        if array_variables_adding_axes_balanced is not None:
            # format axes_inserting_array_not_object
            n_variables_table_adding_axes_balanced = (
                shape_array_variables_adding_axes_balanced[axis_variables_table_input])
            n_axes_inserting_array_not_object = n_variables_table_adding_axes_balanced
            n_axes_array_not_object = n_axes_array_variables_staying + n_axes_inserting_array_not_object
            try:
                len(axes_inserting_array_not_object)
                axes_inserting_array_not_object = np.asarray(axes_inserting_array_not_object, dtype=int)
                axes_inserting_array_not_object[axes_inserting_array_not_object < 0] += n_axes_array_not_object
                # check point 1
                for a in axes_inserting_array_not_object:
                    if np.sum(a == axes_inserting_array_not_object) > 1:
                        raise ValueError('axes_inserting_array_not_object cannot contain repeated values')
            except TypeError:
                if axes_inserting_array_not_object is None:
                    axes_inserting_array_not_object = np.arange(n_axes_inserting_array_not_object)
                else:
                    if axes_inserting_array_not_object < 0:
                        axes_inserting_array_not_object += n_axes_array_not_object
                    axes_inserting_array_not_object = np.asarray([axes_inserting_array_not_object], dtype=int)

        if array_variables_adding_axes_unbalanced is not None:
            # format axes_inserting_array_object
            n_variables_table_adding_axes_unbalanced = (
                shape_array_variables_adding_axes_balanced[axis_variables_table_input])
            n_axes_inserting_array_object = n_variables_table_adding_axes_unbalanced
            n_axes_array_object = n_axes_inserting_array_object
            try:
                len(axes_inserting_array_object)
                axes_inserting_array_object = np.asarray(axes_inserting_array_object, dtype=int)
                axes_inserting_array_object[axes_inserting_array_object < 0] += n_axes_array_object
                # check point 2
                for a in axes_inserting_array_object[0]:
                    if np.sum(a == axes_inserting_array_object) > 1:
                        raise ValueError('axes_inserting_array_object cannot contain repeated values')
            except TypeError:
                if axes_inserting_array_object is None:
                    axes_inserting_array_object = np.arange(n_axes_inserting_array_object)
                else:
                    if axes_inserting_array_object < 0:
                        axes_inserting_array_object += n_axes_array_object
                    axes_inserting_array_object = np.asarray([axes_inserting_array_object], dtype=int)

    if array_variables_adding_axes_balanced is not None and array_variables_adding_axes_unbalanced is not None:

        array_variables_staying = v2a_from_2_arrays(
            array_variables_staying,
            array_variables_adding_axes_balanced,
            axis_samples_input,
            axis_variables_table_input,
            axes_inserting=axes_inserting_array_not_object,
            balance=True,
            dtype=dtype,
            format_and_check=False)

        array_variables_adding_axes_unbalanced = v2a_from_2_arrays(
            array_variables_adding_axes_unbalanced,
            array_variables_adding_axes_balanced,
            axis_samples_input,
            axis_variables_table_input,
            axes_inserting=axes_inserting_array_not_object,
            balance=True,
            dtype=None,
            format_and_check=False)

        axis_variables_table_input += np.sum(axes_inserting_array_not_object <= axis_variables_table_input)
        axis_samples_input += np.sum(axes_inserting_array_not_object <= axis_samples_input)

        array_output = v2a_from_2_arrays(
            array_variables_staying,
            array_variables_adding_axes_unbalanced,
            axis_samples_input,
            axis_variables_table_input,
            axes_inserting=axes_inserting_array_object,
            balance=False,
            dtype=None,
            format_and_check=False)

    elif array_variables_adding_axes_balanced is not None and array_variables_adding_axes_unbalanced is None:

        array_output = v2a_from_2_arrays(
            array_variables_staying,
            array_variables_adding_axes_balanced,
            axis_samples_input,
            axis_variables_table_input,
            axes_inserting=axes_inserting_array_not_object,
            balance=True,
            dtype=dtype,
            format_and_check=False)

    elif array_variables_adding_axes_balanced is None and array_variables_adding_axes_unbalanced is not None:

        if dtype is not None:
            array_variables_staying = array_variables_staying.astype(dtype)

        array_output = v2a_from_2_arrays(
            array_variables_staying,
            array_variables_adding_axes_unbalanced,
            axis_samples_input,
            axis_variables_table_input,
            axes_inserting=axes_inserting_array_object,
            balance=False,
            dtype=None,
            format_and_check=False)

    return array_output


def from_3_arrays_advanced(
        array_variables_staying,
        array_variables_adding_axes_balanced,
        array_variables_adding_axes_unbalanced,
        axis_samples,
        axis_variables_table,
        variables_table_adding_axes_balanced=None,
        variables_table_adding_axes_unbalanced=None,
        axes_inserting_array_not_object=None,
        axes_inserting_array_object=None,
        variables_table_staying=None,
        dtype=None,
        format_and_check=True):

    # Notes:
    # 1) it is safe;
    # 2) it does not assumes that the numbers of samples (or trials) in the axis_samples for all
    #    combinations of variables_table_adding_axes_balanced are equal;
    # 3) it does not assumes that the order of samples (or trials) in the axis_samples are is the same
    #    in each combination of variables' axes. The order is based on the order of the combinations
    #    of variables' conditions in axis_variables_table. In other words, the samples can be in random order
    #    (or in different order) in each combinations of variables axes/conditions.

    # Input requirements:
    # 1) axes_inserting_array_not_object cannot contain repeated values;
    # 2) axes_inserting_array_object cannot contain repeated values;
    # 3) variables_table_adding_axes_balanced cannot contain repeated values;
    # 4) variables_table_adding_axes_unbalanced cannot contain repeated values;
    # 5) shapes of axes_inserting_array_not_object and variables_table_adding_axes_balanced must be equal
    # 6) shapes of axes_inserting_array_object and variables_table_adding_axes_unbalanced must be equal
    # 7) axis_samples != axis_variables_table
    # 8) Any value in variables_table_adding_axes_balanced must be different to
    #    any value in variables_table_adding_axes_unbalanced;
    # 9) Either one of the 3 conditions must be True:
    #    variables_table_adding_axes_balanced != None;
    #    variables_table_adding_axes_unbalanced != None
    #    variables_table_adding_axes_balanced != None and variables_table_adding_axes_unbalanced != None;
    # 10) the numbers of axes of the 3 arrays must be equal;
    # 11) the shapes of the 3 arrays must be equal, except for the axis_variable_table,
    #     i.e. the axis that contains the variables of the tables;

    # Import requirements:
    # 1) import numpy as np
    # 2) from ccalafiore.combinations import n_conditions_to_combinations
    # 3) from ccalafiore.combinations import trials_to_conditions
    # 4) from ccalafiore.combinations import conditions_to_n_conditions
    # 5) from ccalafiore.combinations import conditions_to_combinations
    # 6) from ccalafiore.array import samples_in_arr1_are_in_arr2

    # from ccalafiore.combinations import \
    #     trials_to_conditions, conditions_to_n_conditions, n_conditions_to_combinations, conditions_to_combinations
    # from ccalafiore.array import samples_in_arr1_are_in_arr2

    # check point 9
    if variables_table_adding_axes_balanced is None and variables_table_adding_axes_unbalanced is None:
        raise ValueError('Either one of the 3 conditions must be True:'
                         'variables_table_adding_axes_balanced != None\n'
                         'variables_table_adding_axes_unbalanced != None\n'
                         'variables_table_adding_axes_balanced != None and '
                         'variables_table_adding_axes_unbalanced != None;')

    axis_variables_table_input = axis_variables_table
    axis_samples_input = axis_samples

    shape_array_variables_staying = np.asarray(array_variables_staying.shape)
    n_axes_array_variables_staying = shape_array_variables_staying.size
    n_variables_table_in_array_variables_staying = shape_array_variables_staying[axis_variables_table_input]
    variables_table_in_array_variables_staying = np.arange(n_variables_table_in_array_variables_staying)

    shape_array_variables_adding_axes_balanced = np.asarray(array_variables_adding_axes_balanced.shape)
    n_axes_array_variables_adding_axes_balanced = shape_array_variables_adding_axes_balanced.size
    n_variables_table_in_array_variables_adding_axes_balanced = (
        shape_array_variables_adding_axes_balanced[axis_variables_table_input])
    variables_table_in_array_variables_adding_axes_balanced = np.arange(
        n_variables_table_in_array_variables_adding_axes_balanced)

    shape_array_variables_adding_axes_unbalanced = np.asarray(array_variables_adding_axes_unbalanced.shape)
    n_axes_array_variables_adding_axes_unbalanced = shape_array_variables_adding_axes_unbalanced.size
    n_variables_table_in_array_variables_adding_axes_unbalanced = (
        shape_array_variables_adding_axes_unbalanced[axis_variables_table_input])
    variables_table_in_array_variables_adding_axes_unbalanced = np.arange(
        n_variables_table_in_array_variables_adding_axes_unbalanced)

    if format_and_check:
        # check point 10
        if (n_axes_array_variables_staying != n_axes_array_variables_adding_axes_balanced) or (
                n_axes_array_variables_staying != n_axes_array_variables_adding_axes_unbalanced):
            raise ValueError('dimension mismatch')
        else:
            n_axes_array_input = n_axes_array_variables_staying

        if axis_variables_table_input < 0:
            axis_variables_table_input += n_axes_array_input
        if axis_samples_input < 0:
            axis_samples_input += n_axes_array_input

        # check point 7
        if axis_samples_input == axis_variables_table_input:
            raise ValueError('axis_samples_input and axis_variables_table_input must be different')

        # check point 11
        indexes_logical = np.arange(n_axes_array_input) != axis_variables_table_input
        if np.any(shape_array_variables_staying[indexes_logical] !=
                  shape_array_variables_adding_axes_balanced[indexes_logical]):
            raise ValueError('dimension mismatch')
        if np.any(shape_array_variables_staying[indexes_logical] !=
                  shape_array_variables_adding_axes_unbalanced[indexes_logical]):
            raise ValueError('dimension mismatch')

        # format variables_table_staying
        try:
            n_variables_table_staying = len(variables_table_staying)
            variables_table_staying = np.asarray(variables_table_staying, dtype=int)
            variables_table_staying[variables_table_staying < 0] += n_variables_table_in_array_variables_staying
        except TypeError:
            if variables_table_staying is None:
                variables_table_staying = variables_table_in_array_variables_staying
                n_variables_table_staying = n_variables_table_in_array_variables_staying
            else:
                if variables_table_staying < 0:
                    variables_table_staying += n_variables_table_in_array_variables_staying
                variables_table_staying = np.asarray([variables_table_staying], dtype=int)
                n_variables_table_staying = 1

        if variables_table_adding_axes_balanced is not None:
            # format variables_table_adding_axes_balanced
            try:
                n_variables_table_adding_axes_balanced = len(variables_table_adding_axes_balanced)
                variables_table_adding_axes_balanced = np.asarray(variables_table_adding_axes_balanced, dtype=int)
                variables_table_adding_axes_balanced[variables_table_adding_axes_balanced < 0] += (
                    n_variables_table_in_array_variables_adding_axes_balanced)
                # check point 3
                for v in variables_table_adding_axes_balanced:
                    if np.sum(v == variables_table_adding_axes_balanced) > 1:
                        raise ValueError('variables_table_adding_axes_balanced cannot contain repeated values')
            except TypeError:
                if variables_table_adding_axes_balanced < 0:
                    variables_table_adding_axes_balanced += n_variables_table_in_array_variables_adding_axes_balanced
                variables_table_adding_axes_balanced = np.asarray([variables_table_adding_axes_balanced], dtype=int)
                n_variables_table_adding_axes_balanced = 1

            # format axes_inserting_array_not_object
            n_axes_inserting_array_not_object = n_variables_table_adding_axes_balanced
            n_axes_array_not_object = n_axes_array_variables_staying + n_axes_inserting_array_not_object
            try:
                n_axes_inserting_array_not_object = len(axes_inserting_array_not_object)
                axes_inserting_array_not_object = np.asarray(axes_inserting_array_not_object, dtype=int)
                axes_inserting_array_not_object[axes_inserting_array_not_object < 0] += n_axes_array_not_object
                # check point 1
                for a in axes_inserting_array_not_object:
                    if np.sum(a == axes_inserting_array_not_object) > 1:
                        raise ValueError('axes_inserting_array_not_object cannot contain repeated values')
                # check point 5
                if n_variables_table_adding_axes_balanced != n_axes_inserting_array_not_object:
                    raise ValueError('Shapes of axes_inserting_array_not_object and '
                                     'variables_table_adding_axes_balanced must be equal')
            except TypeError:
                if axes_inserting_array_not_object is None:
                    axes_inserting_array_not_object = np.arange(n_axes_inserting_array_not_object)
                else:
                    if axes_inserting_array_not_object < 0:
                        axes_inserting_array_not_object += n_axes_array_not_object
                    axes_inserting_array_not_object = np.asarray([axes_inserting_array_not_object], dtype=int)
                    n_axes_inserting_array_not_object = 1
                    # check point 5
                    if n_variables_table_adding_axes_balanced != n_axes_inserting_array_not_object:
                        raise ValueError('Shapes of axes_inserting_array_not_object and '
                                         'variables_table_adding_axes_balanced must be equal')

        if variables_table_adding_axes_unbalanced is not None:
            # format variables_table_adding_axes_unbalanced
            try:
                n_variables_table_adding_axes_unbalanced = len(variables_table_adding_axes_unbalanced)
                variables_table_adding_axes_unbalanced = np.asarray(variables_table_adding_axes_unbalanced, dtype=int)
                variables_table_adding_axes_unbalanced[variables_table_adding_axes_unbalanced < 0] += (
                    n_variables_table_in_array_variables_adding_axes_balanced)
                # check point 4
                for v in variables_table_adding_axes_unbalanced:
                    if np.sum(v == variables_table_adding_axes_unbalanced) > 1:
                        raise ValueError('variables_table_adding_axes_unbalanced cannot contain repeated values')
            except TypeError:
                if variables_table_adding_axes_unbalanced < 0:
                    variables_table_adding_axes_unbalanced += n_variables_table_in_array_variables_adding_axes_balanced
                variables_table_adding_axes_unbalanced = np.asarray([variables_table_adding_axes_unbalanced], dtype=int)
                n_variables_table_adding_axes_unbalanced = 1

            # format axes_inserting_array_object
            n_axes_inserting_array_object = n_variables_table_adding_axes_unbalanced
            n_axes_array_object = n_axes_inserting_array_object
            try:
                n_axes_inserting_array_object = len(axes_inserting_array_object)
                axes_inserting_array_object = np.asarray(axes_inserting_array_object, dtype=int)
                axes_inserting_array_object[axes_inserting_array_object < 0] += n_axes_array_object
                # check point 2
                for a in axes_inserting_array_object:
                    if np.sum(a == axes_inserting_array_object) > 1:
                        raise ValueError('axes_inserting_array_object cannot contain repeated values')
                # check point 6
                if n_variables_table_adding_axes_unbalanced != n_axes_inserting_array_object:
                    raise ValueError('Shapes of axes_inserting_array_object and '
                                     'variables_table_adding_axes_unbalanced must be equal')
            except TypeError:
                if axes_inserting_array_object is None:
                    axes_inserting_array_object = np.arange(n_axes_inserting_array_object)
                else:
                    if axes_inserting_array_object < 0:
                        axes_inserting_array_object += n_axes_array_object
                    axes_inserting_array_object = np.asarray([axes_inserting_array_object], dtype=int)
                    n_axes_inserting_array_object = 1
                    # check point 6
                    if n_variables_table_adding_axes_unbalanced != n_axes_inserting_array_object:
                        raise ValueError('Shapes of axes_inserting_array_object and '
                                         'variables_table_adding_axes_unbalanced must be equal')
    else:
        n_variables_table_staying = variables_table_staying.size
        if variables_table_adding_axes_balanced is not None:
            n_variables_table_adding_axes_balanced = variables_table_adding_axes_balanced.size
        if variables_table_adding_axes_unbalanced is not None:
            n_variables_table_adding_axes_unbalanced = variables_table_adding_axes_unbalanced.size

    indexes = None
    if ((n_variables_table_staying != n_variables_table_in_array_variables_staying) or
            np.any(variables_table_staying != variables_table_in_array_variables_staying)):

        indexes = np.empty(n_axes_array_variables_staying, dtype=object)
        indexes[:] = slice(None)
        indexes[axis_variables_table_input] = variables_table_staying
        array_variables_staying = array_variables_staying[tuple(indexes)]

    if variables_table_adding_axes_balanced is not None:

        if ((n_variables_table_adding_axes_balanced != n_variables_table_in_array_variables_adding_axes_balanced)
                or np.any(variables_table_adding_axes_balanced !=
                          variables_table_in_array_variables_adding_axes_balanced)):

            if indexes is None:
                indexes = np.empty(n_axes_array_variables_adding_axes_balanced, dtype=object)
                indexes[:] = slice(None)
            indexes[axis_variables_table_input] = variables_table_adding_axes_balanced
            array_variables_adding_axes_balanced = array_variables_adding_axes_balanced[tuple(indexes)]

    if variables_table_adding_axes_unbalanced is not None:

        if ((n_variables_table_adding_axes_unbalanced != n_variables_table_in_array_variables_adding_axes_unbalanced)
                or np.any(variables_table_adding_axes_unbalanced !=
                          variables_table_in_array_variables_adding_axes_unbalanced)):
            if indexes is None:
                indexes = np.empty(n_axes_array_variables_adding_axes_unbalanced, dtype=object)
                indexes[:] = slice(None)
            indexes[axis_variables_table_input] = variables_table_adding_axes_unbalanced
            array_variables_adding_axes_unbalanced = array_variables_adding_axes_unbalanced[tuple(indexes)]

    if variables_table_adding_axes_balanced is not None and variables_table_adding_axes_unbalanced is not None:
        # from ccalafiore.array import samples_in_arr1_are_in_arr2
        # check point 8
        if np.any(samples_in_arr1_are_in_arr2(
                variables_table_adding_axes_balanced, variables_table_adding_axes_unbalanced)):
            raise ValueError('Any value in variables_table_adding_axes_balanced must be different to\n'
                             'any value in variables_table_adding_axes_unbalanced')

        array_variables_staying = v2a_from_2_arrays(
            array_variables_staying,
            array_variables_adding_axes_balanced,
            axis_samples_input,
            axis_variables_table_input,
            axes_inserting_output=axes_inserting_array_not_object,
            balance=True,
            dtype=dtype,
            format_and_check=False)

        array_variables_adding_axes_unbalanced = v2a_from_2_arrays(
            array_variables_adding_axes_unbalanced,
            array_variables_adding_axes_balanced,
            axis_samples_input,
            axis_variables_table_input,
            axes_inserting_output=axes_inserting_array_not_object,
            balance=True,
            dtype=None,
            format_and_check=False)

        axis_variables_table_input += np.sum(axes_inserting_array_not_object <= axis_variables_table_input)
        axis_samples_input += np.sum(axes_inserting_array_not_object <= axis_samples_input)

        array_output = v2a_from_2_arrays(
            array_variables_staying,
            array_variables_adding_axes_unbalanced,
            axis_samples_input,
            axis_variables_table_input,
            axes_inserting_output=axes_inserting_array_object,
            balance=False,
            dtype=None,
            format_and_check=False)

    elif variables_table_adding_axes_balanced is not None and variables_table_adding_axes_unbalanced is None:

        array_output = v2a_from_2_arrays(
            array_variables_staying,
            array_variables_adding_axes_balanced,
            axis_samples_input,
            axis_variables_table_input,
            axes_inserting=axes_inserting_array_not_object,
            balance=True,
            dtype=dtype,
            format_and_check=False)

    elif variables_table_adding_axes_balanced is None and variables_table_adding_axes_unbalanced is not None:

        if dtype is not None:
            array_variables_staying = array_variables_staying.astype(dtype)

        array_output = v2a_from_2_arrays(
            array_variables_staying,
            array_variables_adding_axes_unbalanced,
            axis_samples_input,
            axis_variables_table_input,
            axes_inserting=axes_inserting_array_object,
            balance=False,
            dtype=None,
            format_and_check=False)

    return array_output


def from_1_array(
        array_input,
        axis_samples,
        axis_variables_table,
        variables_table_adding_axes_balanced=None,
        variables_table_adding_axes_unbalanced=None,
        axes_inserting_array_not_object=None,
        axes_inserting_array_object=None,
        variables_table_staying=None,
        dtype=None,
        format_and_check=True):

    # Notes:
    # 1) it is safe;
    # 2) it does not assumes that the numbers of samples (or trials) in the axis_samples for all
    #    combinations of variables_table_adding_axes_balanced are equal;
    # 3) it does not assumes that the order of samples (or trials) in the axis_samples are is the same
    #    in each combination of variables' axes. The order is based on the order of the combinations
    #    of variables' conditions in axis_variables_table. In other words, the samples can be in random order
    #    (or in different order) in each combinations of variables axes/conditions.

    # Input requirements:
    # 1) axes_inserting_array_not_object cannot contain same values;
    # 2) axes_inserting_array_object cannot contain same values;
    # 3) variables_table_adding_axes_balanced cannot contain same values;
    # 4) variables_table_adding_axes_unbalanced cannot contain same values;
    # 5) shapes of axes_inserting_array_not_object and variables_table_adding_axes_balanced must be equal
    # 6) shapes of axes_inserting_array_object and variables_table_adding_axes_unbalanced must be equal
    # 7) axis_samples != axis_variables_table
    # 8) Any value in variables_table_adding_axes_balanced must be different to
    #    any value in variables_table_adding_axes_unbalanced;
    # 9) Either one of the 3 conditions must be True:
    #    variables_table_adding_axes_balanced != None;
    #    variables_table_adding_axes_unbalanced != None
    #    variables_table_adding_axes_balanced != None and variables_table_adding_axes_unbalanced != None;'

    # Import requirements:
    # 1) import numpy as np
    # 2) from ccalafiore.combinations import n_conditions_to_combinations
    # 3) from ccalafiore.combinations import trials_to_conditions
    # 4) from ccalafiore.combinations import conditions_to_n_conditions
    # 5) from ccalafiore.combinations import conditions_to_combinations
    # 6) from ccalafiore.array import samples_in_arr1_are_in_arr2

    # from ccalafiore.combinations import \
    #     trials_to_conditions, conditions_to_n_conditions, n_conditions_to_combinations, conditions_to_combinations
    # from ccalafiore.array import samples_in_arr1_are_in_arr2

    # check point 9
    if variables_table_adding_axes_balanced is None and variables_table_adding_axes_unbalanced is None:
        raise ValueError('Either one of the 3 conditions must be True:'
                         'variables_table_adding_axes_balanced != None\n'
                         'variables_table_adding_axes_unbalanced != None\n'
                         'variables_table_adding_axes_balanced != None and '
                         'variables_table_adding_axes_unbalanced != None;')

    axis_variables_table_input = axis_variables_table
    axis_samples_input = axis_samples

    shape_array_input = np.asarray(array_input.shape)
    n_axes_array_input = shape_array_input.size
    n_variables_table_input = shape_array_input[axis_variables_table_input]
    variables_table_input = np.arange(n_variables_table_input)

    if format_and_check:

        if axis_variables_table_input < 0:
            axis_variables_table_input += n_axes_array_input
        if axis_samples_input < 0:
            axis_samples_input += n_axes_array_input

        # check point 7
        if axis_samples_input == axis_variables_table_input:
            raise ValueError('axis_samples_input and axis_variables_table_input must be different')

        variables_table_adding_axes = None

        if variables_table_adding_axes_balanced is not None:

            # format variables_table_adding_axes_balanced
            try:
                n_variables_table_adding_axes_balanced = len(variables_table_adding_axes_balanced)
                variables_table_adding_axes_balanced = np.asarray(variables_table_adding_axes_balanced, dtype=int)
                variables_table_adding_axes_balanced[variables_table_adding_axes_balanced < 0] += (
                    n_variables_table_input)
                # check point 3
                for v in variables_table_adding_axes_balanced:
                    if np.sum(v == variables_table_adding_axes_balanced) > 1:
                        raise ValueError('variables_table_adding_axes_balanced cannot contain repeated values')
            except TypeError:
                if variables_table_adding_axes_balanced < 0:
                    variables_table_adding_axes_balanced += n_variables_table_input
                variables_table_adding_axes_balanced = np.asarray([variables_table_adding_axes_balanced], dtype=int)
                n_variables_table_adding_axes_balanced = 1

            variables_table_adding_axes = variables_table_adding_axes_unbalanced

            # format axes_inserting_array_not_object
            n_axes_inserting_array_not_object = n_variables_table_adding_axes_balanced
            n_axes_array_not_object = n_axes_array_input + n_axes_inserting_array_not_object
            try:
                n_axes_inserting_array_not_object = len(axes_inserting_array_not_object)
                axes_inserting_array_not_object = np.asarray(axes_inserting_array_not_object, dtype=int)
                axes_inserting_array_not_object[axes_inserting_array_not_object < 0] += n_axes_array_not_object
                # check point 1
                for a in axes_inserting_array_not_object:
                    if np.sum(a == axes_inserting_array_not_object) > 1:
                        raise ValueError('axes_inserting_array_not_object cannot contain repeated values')
                # check point 5
                if n_variables_table_adding_axes_balanced != n_axes_inserting_array_not_object:
                    raise ValueError('Shapes of axes_inserting_array_not_object and '
                                     'variables_table_adding_axes_balanced must be equal')
            except TypeError:
                if axes_inserting_array_not_object is None:
                    axes_inserting_array_not_object = np.arange(n_axes_inserting_array_not_object)
                else:
                    if axes_inserting_array_not_object < 0:
                        axes_inserting_array_not_object += n_axes_array_not_object
                    axes_inserting_array_not_object = np.asarray([axes_inserting_array_not_object], dtype=int)
                    n_axes_inserting_array_not_object = 1
                    # check point 5
                    if n_variables_table_adding_axes_balanced != n_axes_inserting_array_not_object:
                        raise ValueError('Shapes of axes_inserting_array_not_object and '
                                         'variables_table_adding_axes_balanced must be equal')

        if variables_table_adding_axes_unbalanced is not None:
            # format variables_table_adding_axes_unbalanced
            try:
                n_variables_table_adding_axes_unbalanced = len(variables_table_adding_axes_unbalanced)
                variables_table_adding_axes_unbalanced = np.asarray(variables_table_adding_axes_unbalanced, dtype=int)
                variables_table_adding_axes_unbalanced[variables_table_adding_axes_unbalanced < 0] += \
                    n_variables_table_input
                # check point 4
                for v in variables_table_adding_axes_unbalanced:
                    if np.sum(v == variables_table_adding_axes_unbalanced) > 1:
                        raise ValueError('variables_table_adding_axes_unbalanced cannot contain repeated values')
            except TypeError:
                if variables_table_adding_axes_unbalanced < 0:
                    variables_table_adding_axes_unbalanced += n_variables_table_input
                variables_table_adding_axes_unbalanced = np.asarray([variables_table_adding_axes_unbalanced], dtype=int)
                n_variables_table_adding_axes_unbalanced = 1

            if variables_table_adding_axes is None:
                variables_table_adding_axes = variables_table_adding_axes_unbalanced
            else:
                variables_table_adding_axes = np.append(
                    variables_table_adding_axes, variables_table_adding_axes_unbalanced, axis=0)

            # format axes_inserting_array_object
            n_axes_inserting_array_object = n_variables_table_adding_axes_unbalanced
            n_axes_array_object = n_axes_inserting_array_object
            try:
                n_axes_inserting_array_object = len(axes_inserting_array_object)
                axes_inserting_array_object = np.asarray(axes_inserting_array_object, dtype=int)
                axes_inserting_array_object[axes_inserting_array_object < 0] += n_axes_array_object
                # check point 2
                for a in axes_inserting_array_object:
                    if np.sum(a == axes_inserting_array_object) > 1:
                        raise ValueError('axes_inserting_array_object cannot contain repeated values')
                # check point 6
                if n_variables_table_adding_axes_unbalanced != n_axes_inserting_array_object:
                    raise ValueError('Shapes of axes_inserting_array_object and '
                                     'variables_table_adding_axes_unbalanced must be equal')
            except TypeError:
                if axes_inserting_array_object is None:
                    axes_inserting_array_object = np.arange(n_axes_inserting_array_object)
                else:
                    if axes_inserting_array_object < 0:
                        axes_inserting_array_object += n_axes_array_object
                    axes_inserting_array_object = np.asarray([axes_inserting_array_object], dtype=int)
                    n_axes_inserting_array_object = 1
                    # check point 6
                    if n_variables_table_adding_axes_unbalanced != n_axes_inserting_array_object:
                        raise ValueError('Shapes of axes_inserting_array_object and '
                                         'variables_table_adding_axes_unbalanced must be equal')

        # format variables_table_staying
        try:
            n_variables_table_staying = len(variables_table_staying)
            variables_table_staying = np.asarray(variables_table_staying, dtype=int)
            variables_table_staying[variables_table_staying < 0] += n_variables_table_input
        except TypeError:
            if variables_table_staying is None:
                variables_table_staying = variables_table_input[np.logical_not(
                    samples_in_arr1_are_in_arr2(variables_table_input, variables_table_adding_axes))]
                n_variables_table_staying = variables_table_staying.size
            else:
                if variables_table_staying < 0:
                    variables_table_staying += n_variables_table_input
                variables_table_staying = np.asarray([variables_table_staying], dtype=int)
                n_variables_table_staying = 1
    else:
        n_variables_table_staying = variables_table_staying.size
        if variables_table_adding_axes_balanced is not None:
            n_variables_table_adding_axes_balanced = variables_table_adding_axes_balanced.size
        if variables_table_adding_axes_unbalanced is not None:
            n_variables_table_adding_axes_unbalanced = variables_table_adding_axes_unbalanced.size

    indexes = np.empty(n_axes_array_input, dtype=object)
    indexes[:] = slice(None)
    if ((n_variables_table_staying != n_variables_table_input) or
            np.any(variables_table_staying != variables_table_input)):
        indexes[axis_variables_table_input] = variables_table_staying
    array_variables_staying = array_input[tuple(indexes)]

    if variables_table_adding_axes_balanced is not None:
        if ((n_variables_table_adding_axes_balanced != n_variables_table_input) or
                np.any(variables_table_adding_axes_balanced != variables_table_input)):

            indexes[axis_variables_table_input] = variables_table_adding_axes_balanced
        else:
            indexes[axis_variables_table_input] = slice(None)
        array_variables_adding_axes_balanced = array_input[tuple(indexes)]

    if variables_table_adding_axes_unbalanced is not None:
        if ((n_variables_table_adding_axes_unbalanced != n_variables_table_input) or
                np.any(variables_table_adding_axes_unbalanced != variables_table_input)):

            indexes[axis_variables_table_input] = variables_table_adding_axes_unbalanced
        else:
            indexes[axis_variables_table_input] = slice(None)
        array_variables_adding_axes_unbalanced = array_input[tuple(indexes)]

    del array_input

    if variables_table_adding_axes_balanced is not None and variables_table_adding_axes_unbalanced is not None:
        # from ccalafiore.array import samples_in_arr1_are_in_arr2
        # check point 8
        if np.any(samples_in_arr1_are_in_arr2(
                variables_table_adding_axes_balanced, variables_table_adding_axes_unbalanced)):
            raise ValueError('Any value in variables_table_adding_axes_balanced must be different to\n'
                             'any value in variables_table_adding_axes_unbalanced')

        # variables_table_staying_not_object = np.unique(np.append(
        #     variables_table_staying, variables_table_adding_axes_unbalanced))

        array_variables_staying = v2a_from_2_arrays(
            array_variables_staying,
            array_variables_adding_axes_balanced,
            axis_samples_input,
            axis_variables_table_input,
            axes_inserting=axes_inserting_array_not_object,
            balance=True,
            dtype=dtype,
            format_and_check=False)

        array_variables_adding_axes_unbalanced = v2a_from_2_arrays(
            array_variables_adding_axes_unbalanced,
            array_variables_adding_axes_balanced,
            axis_samples_input,
            axis_variables_table_input,
            axes_inserting=axes_inserting_array_not_object,
            balance=True,
            dtype=None,
            format_and_check=False)

        # axis_variables_table_input += np.sum(axes_inserting_array_not_object <= axis_variables_table_input)
        # axis_samples_input += np.sum(axes_inserting_array_not_object <= axis_samples_input)
        axis_variables_table_output = axis_variables_table_input + np.sum(
            axes_inserting_array_not_object <= axis_variables_table_input)
        axis_samples_output = axis_samples_input + np.sum(axes_inserting_array_not_object <= axis_samples_input)
        changed = True
        while changed:
            changed = False
            while axis_variables_table_output in axes_inserting_array_not_object:
                axis_variables_table_output += 1
                changed = True
            while axis_samples_output in axes_inserting_array_not_object:
                axis_samples_output += 1
                changed = True
            if axis_variables_table_output == axis_samples_output:
                changed = True
                if axis_variables_table_input > axis_samples_input:
                    axis_variables_table_output += 1
                elif axis_variables_table_input < axis_samples_input:
                    axis_samples_output += 1

        axis_variables_table_input = axis_variables_table_output
        axis_samples_input = axis_samples_output

        array_output = v2a_from_2_arrays(
            array_variables_staying,
            array_variables_adding_axes_unbalanced,
            axis_samples_input,
            axis_variables_table_input,
            axes_inserting=axes_inserting_array_object,
            balance=False,
            dtype=None,
            format_and_check=False)

    elif variables_table_adding_axes_balanced is not None and variables_table_adding_axes_unbalanced is None:

        array_output = v2a_from_2_arrays(
            array_variables_staying,
            array_variables_adding_axes_balanced,
            axis_samples_input,
            axis_variables_table_input,
            axes_inserting=axes_inserting_array_not_object,
            balance=True,
            dtype=dtype,
            format_and_check=False)

    elif variables_table_adding_axes_balanced is None and variables_table_adding_axes_unbalanced is not None:

        if dtype is not None:
            array_variables_staying = array_variables_staying.astype(dtype)

        array_output = v2a_from_2_arrays(
            array_variables_staying,
            array_variables_adding_axes_unbalanced,
            axis_samples_input,
            axis_variables_table_input,
            axes_inserting=axes_inserting_array_object,
            balance=False,
            dtype=None,
            format_and_check=False)

    return array_output



import numpy as np
from copy import deepcopy
from . import descriptive
from .. import array as cc_array
from .. import maths as cc_maths
from .. import format as cc_format


def exclude_trials_with_rt_out_of_range(
        data, axes, indexes,
        lower_limit, upper_limit, value=np.nan, copy=False):

    try:
        axes[0]
        axes_ = axes
    except TypeError:
        axes_ = [axes]

    try:
        indexes[0]
        indexes_ = indexes
    except TypeError:
        indexes_ = [indexes]

    condition = np.empty(data.shape, dtype=bool)

    n_axes = len(data.shape)
    indexes_data = np.empty(n_axes, dtype='O')
    indexes_condition = np.empty(n_axes, dtype='O')
    j = 0
    for a in range(n_axes):
        if a in axes_:
            try:
                indexes_[j][0]
            except TypeError:
                if not isinstance(indexes_[j], slice):
                    indexes_[j] = slice(indexes_[j], indexes_[j] + 1, 1)

            indexes_data[a] = indexes_[j]
            j += 1
        else:
            indexes_data[a] = slice(0, data.shape[a], 1)

        indexes_condition[a] = slice(0, data.shape[a], 1)

    data_to_check = data[tuple(indexes_data)]

    condition[tuple(indexes_condition)] = np.logical_or(data_to_check < lower_limit, upper_limit < data_to_check)

    if copy:
        data_out = deepcopy(data)
        data_out[condition] = value
        return data_out
    else:
        data[condition] = value
        return data


def exclude_subs_with_lower_accuracy(corrects, axis, threshold, order='n', exclude_nan=True):
    """

    Parameters
    ----------
    corrects
    axis
    threshold
    order: str, optional
        The desired outputs and their order. Accepted values are "n", "l", "a" or any combination of them like "al",
        "lan" (default is "n").
    exclude_nan

    Returns
    -------

    """
    n_outputs = len(order)
    outputs = [None] * n_outputs  # type: list

    for o in range(n_outputs):
        if order[o] in 'nla':

            accuracies_subs = descriptive.scores_to_means(corrects, axes=axis, exclude_nan=exclude_nan)

            if 'a' in order:
                outputs[order.index('a')] = accuracies_subs
            break

    for o in range(n_outputs):
        if order[o] in 'nl':
            subs_excluded_logical = accuracies_subs < threshold
            if 'l' in order:
                outputs[order.index('l')] = subs_excluded_logical
            break

    for o in range(n_outputs):
        if order[o] in 'n':
            subs_excluded_numeric = np.where(subs_excluded_logical)[0]
            if 'n' in order:
                outputs[order.index('n')] = subs_excluded_numeric
            break

    return outputs


def exclude_subs_with_lower_local_accuracy(corrects, axis, n_neighbours, threshold, order='n', exclude_nan=False):
    """

    Parameters
    ----------
    corrects
    axis
    n_neighbours
    threshold
    order: str, optional
        The desired outputs and their order. Accepted values are "n", "l", "m" or any combination of them like "ml",
        "lmn" (default is "n").
    exclude_nan

    Returns
    -------

    """
    n_outputs = len(order)
    outputs = [None] * n_outputs  # type: list

    for o in range(n_outputs):
        if order[o] in 'nlm':
            accuracy_local = descriptive.scores_to_local_mean(corrects, axis, n_neighbours, exclude_nan=exclude_nan)
            min_accuracy_local = accuracy_local.min(axis=axis, initial=1)
            if 'm' in order:
                outputs[order.index('m')] = min_accuracy_local
            break

    for o in range(n_outputs):
        if order[o] in 'nl':
            subs_excluded_logical = min_accuracy_local < threshold
            if 'l' in order:
                outputs[order.index('l')] = subs_excluded_logical
            break

    for o in range(n_outputs):
        if order[o] in 'n':
            subs_excluded_numeric = np.where(subs_excluded_logical)[0]
            if 'n' in order:
                outputs[order.index('n')] = subs_excluded_numeric
            break

    return outputs


def exclude_subs_with_higher_excluded_trials(data, axis, threshold, percentage=False, value=np.nan, order='n'):
    """

    Parameters
    ----------
    data : np.ndarray
        Some data with size S x N, where S is the number of subjects and N is the number of trials per subject.
        The excluded trials in the data were flagged with a specific value that you need to define in the
        parameter "value".
    axis : int or sequence or ints
        The axis of the data of the trials.
    threshold : int, float
        exclude subjects with excluded trials greater than threshold
    percentage : bool, optional
        If percentage is True, define threshold in percentage. Else, define threshold in number excluded trials.
    value : Any, optional
        The value that was used to flag the excluded trials in the data.
    order: str, optional
        The desired outputs and their order. Accepted values are "n", "l", "e" or any combination of them like "el",
        "len" (default is "n").

    Returns
    -------
        outputs : list
            A list of the elements requested in the parameter "order".

    """

    n_outputs = len(order)
    outputs = [None] * n_outputs  # type: list

    for o in range(n_outputs):
        if order[o] in 'nle':
            n_excluded_trials_subs = descriptive.scores_to_value_frequencies(data, axis, value)
            # if cc_maths.is_nan(value):
            #     excluded_trials_logical = cc_maths.is_nan(data)
            # else:
            #     excluded_trials_logical = data == value
            # n_excluded_trials_subs = np.sum(excluded_trials_logical, axis=axis, initial=0)
            break

    for o in range(n_outputs):
        if order[o] in 'nl':
            if percentage:
                N = data.shape[axis]

                if 'e' in order:
                    p_threshold = threshold
                    p_excluded_trials_subs = n_excluded_trials_subs * (100 / N)
                    subs_excluded_logical = p_excluded_trials_subs > p_threshold
                    outputs[order.index('e')] = p_excluded_trials_subs
                else:
                    n_threshold = threshold * (N / 100)
                    subs_excluded_logical = n_excluded_trials_subs > n_threshold
            else:
                n_threshold = threshold
                subs_excluded_logical = n_excluded_trials_subs > n_threshold

                if 'e' in order:
                    outputs[order.index('e')] = n_excluded_trials_subs

            if 'l' in order:
                outputs[order.index('l')] = subs_excluded_logical
            break

    for o in range(n_outputs):
        if order[o] in 'n':
            subs_excluded_numeric = np.where(subs_excluded_logical)[0]
            if 'n' in order:
                outputs[order.index('n')] = subs_excluded_numeric
            break

    return outputs


def exclude_subs_with_higher_local_excluded_trials(
        data, axis, n_neighbours, threshold, percentage=False, value=np.nan, order='n'):
    """

    Parameters
    ----------
    data : np.ndarray
        Some data with size S x N, where S is the number of subjects and N is the number of trials per subject.
        The excluded trials in the data were flagged with a specific value that you need to define in the
        parameter "value".
    axis : int or sequence or ints
        The axis of the data of the trials.
    threshold : int, float
        exclude subjects with excluded trials greater than threshold
    n_neighbours : int
        to add
    percentage : bool, optional
        If percentage is True, define threshold in percentage. Else, define threshold in number excluded trials.
    value : Any, optional
        The value that was used to flag the excluded trials in the data.
    order: str, optional
        The desired outputs and their order. Accepted values are "n", "l", "e" or any combination of them like "el",
        "len" (default is "n").

    Returns
    -------
        outputs : list
            A list of the elements requested in the parameter "order".

    """

    n_outputs = len(order)
    outputs = [None] * n_outputs  # type: list

    for o in range(n_outputs):
        if order[o] in 'nle':
            local_value_frequencies = descriptive.scores_to_local_value_frequencies(data, axis, n_neighbours, value)
            local_n_excluded_trials_subs = np.max(local_value_frequencies, axis=axis, initial=0)
            break

    for o in range(n_outputs):
        if order[o] in 'nl':
            if percentage:
                N = data.shape[axis]
                # N = n_neighbours

                if 'e' in order:
                    p_threshold = threshold
                    local_p_excluded_trials_subs = local_n_excluded_trials_subs * (100 / N)
                    subs_excluded_logical = local_p_excluded_trials_subs > p_threshold
                    outputs[order.index('e')] = local_p_excluded_trials_subs
                else:
                    n_threshold = threshold * (N / 100)
                    subs_excluded_logical = local_n_excluded_trials_subs > n_threshold
            else:
                n_threshold = threshold
                subs_excluded_logical = local_n_excluded_trials_subs > n_threshold

                if 'e' in order:
                    outputs[order.index('e')] = local_n_excluded_trials_subs

            if 'l' in order:
                outputs[order.index('l')] = subs_excluded_logical
            break

    for o in range(n_outputs):
        if order[o] in 'n':
            subs_excluded_numeric = np.where(subs_excluded_logical)[0]
            if 'n' in order:
                outputs[order.index('n')] = subs_excluded_numeric
            break

    return outputs


def exclude_subs(data, axis, subs_exc):

    subs_all = np.arange(data.shape[axis])
    subs_keep = subs_all[cc_array.samples_in_arr1_are_not_in_arr2(subs_all, subs_exc, axis=0)]

    n_axes = len(data.shape)
    indexes_data = np.empty(n_axes, dtype='O')
    for a in range(n_axes):
        if a == axis:
            indexes_data[a] = cc_format.numeric_indexes_to_slice(subs_keep)
        else:
            indexes_data[a] = slice(0, data.shape[a], 1)

    data_out = data[tuple(indexes_data)]

    return data_out
